"""Create Haskell style do blocks from generator function
"""
import inspect
from functools import wraps
from .result import Result

__all__ = ('do', 'do_return', 'do_done',)


def do(Monad):
    """Do block

    Create Haskell style do blocks from generator function. One obvious
    limitation of this implementation it cannot be used with monad which call
    bound function multiple times inside bind implementation (e.g. List monad).
    """
    unit = Monad.unit
    error = lambda: unit(Result.from_current_error())
    value = lambda val: unit(Result.from_value(val))

    def do(block):
        if not inspect.isgeneratorfunction(block):
            @wraps(block)
            def do_block(*a, **kw):
                try:
                    return value(block(*a, **kw))
                except _return as ret:
                    return value(ret.args[1]) if ret.args[0] == 0 else ret.args[1]
                except Exception:
                    return error()
        else:
            @wraps(block)
            def do_block(*args, **kw):
                def do_next(result):
                    val, err = (result.pair if isinstance(result, Result) else
                               (result, None))
                    try:
                        monad = (gen.send(val) if err is None else
                                 gen.throw(*err)).__monad__()
                        return monad.bind(do_next)
                    except _return as ret:
                        gen.close()
                        return value(ret.args[1]) if ret.args[0] == 0 else ret.args[1]
                    except StopIteration as ret:
                        gen.close()
                        return value(ret.args[0] if ret.args else None)
                    except Exception:
                        return error()

                try:
                    gen = block(*args, **kw)
                    return do_next(Result.from_value(None))
                except Exception:
                    return error()
        return do_block
    return do


def do_return(value):
    """Return value from do block
    """
    raise _return(0, value)


def do_done(monad):
    """Return monad from do block
    """
    raise _return(1, monad)


class _return(BaseException):
    """Helper exception used to return from do block

    Use separate result class instead of StopIteration because otherwise it will
    be handled in standard exception closure.
    """



import os
import unittest
from ..conn.composite import composite_fork_conn
from ...tests import async_test

__all__ = ('CompositeConnectionTest',)


class CompositeConnectionTest(unittest.TestCase):
    @async_test
    def test_flat(self):
        with (yield composite_fork_conn(3, mesh='flat')) as conns:
            pid = os.getpid()
            for ppid in (yield conns(os.getppid)()):
                self.assertEqual(pid, ppid)

    @async_test
    def test_mesh(self):
        with (yield composite_fork_conn(3, mesh='tree:2')) as conns:
            pids = dict(zip((yield conns(os.getpid)()),
                            (yield conns(os.getppid)())))
        direct = list(pid for pid, ppid in pids.items() if ppid == os.getpid())
        indirect = list(pid for pid, ppid in pids.items() if ppid in direct)
        self.assertEqual(len(direct), 2)
        self.assertEqual(len(indirect), 1)



# -*- coding: utf-8 -*-
import zlib
import json
from .. import PRETZEL_BUFSIZE

__all__ = ('StoreStream',)


class StoreStream(object):
    """Stream object with Store back-end.
    """
    default_compress = 9

    def __init__(self, store, header, bufsize=None, compress=None):
        self.store = store
        self.header = header

        header = self.header()
        if not header:
            self.bufsize = bufsize or PRETZEL_BUFSIZE
            self.chunks = []
            self.size = 0
            self.compress = self.default_compress if compress is None else compress
        else:
            header = json.loads(header.decode())
            self.bufsize = header['bufsize']
            self.chunks = header['chunks']
            self.size = header['size']
            self.compress = header['compress']

        self.seek_pos = None
        self.chunk_index = None
        self.chunk_dirty = False
        self.chunk_zero = b'\x00' * self.bufsize
        self.chunk_switch(0)

    def chunk_switch(self, index=None):
        """Switch current chunk
        """
        if self.chunk_index == index:
            return

        if self.chunk_dirty:
            self.chunk_dirty = False
            chunk_desc = self.store.save(
                self.chunk.bytes() if not self.compress else
                zlib.compress(self.chunk.bytes(), self.compress))
            if self.chunk_index < len(self.chunks):
                self.chunks[self.chunk_index] = chunk_desc
            else:
                self.chunks.append(chunk_desc)

        self.chunk_index = self.chunk_index + 1 if index is None else index
        if self.chunk_index < len(self.chunks):
            self.chunk_desc = self.chunks[self.chunk_index]
            self.chunk = Chunk(
                self.bufsize,
                self.chunk_zero if self.chunk_desc is None else
                self.store.load(self.chunk_desc) if not self.compress else
                zlib.decompress(self.store.load(self.chunk_desc)))
        else:
            self.chunks.extend((None,) * (self.chunk_index - len(self.chunks)))
            self.chunk_desc = None
            self.chunk = Chunk(self.bufsize)

    def write(self, data):
        """Write data to stream
        """
        if self.seek_pos is not None:
            self.seek_do()

        data_size = len(data)
        data_offset = 0
        while True:
            data_offset += self.chunk.write(data[data_offset:])
            self.chunk_dirty = True
            if data_offset == data_size:
                break
            self.chunk_switch()

        self.size = max(self.size, self.chunk_index * self.bufsize + self.chunk.tell())
        return len(data)

    def read(self, size=None):
        """Read data from stream
        """
        if self.seek_pos is not None:
            if self.seek_pos < self.size:
                self.seek_do()
            else:
                return b''

        size = self.size if size is None else size
        data = []
        data_size = 0

        while True:
            chunk = self.chunk.read(size - data_size)
            data.append(chunk)
            data_size += len(chunk)
            if data_size == size:
                break
            if len(self.chunks) <= self.chunk_index + 1:
                break
            else:
                self.chunk_switch()

        return b''.join(data)

    def seek(self, pos, whence=0):
        """Seek stream
        """
        if whence == 0:    # SEEK_SET
            self.seek_pos = pos
        elif whence == 1:  # SEEK_CUR
            self.seek_pos = self.chunk_index * self.bufsize + self.chunk.tell() + pos
        elif whence == 2:  # SEEK_END
            self.seek_pos = self.size + pos
        else:
            raise ValueError('Invalid whence argument: {}'.format(whence))
        return self.seek_pos

    def seek_do(self):
        """Actually seek
        """
        seek_pos, self.seek_pos = self.seek_pos, None
        if seek_pos is None:
            return
        index, offset = divmod(seek_pos, self.bufsize)
        self.chunk_switch(index)
        self.chunk.seek(offset)

    def tell(self):
        """Tell current position inside stream
        """
        if self.seek_pos is None:
            return self.chunk_index * self.bufsize + self.chunk.tell()
        else:
            return self.seek_pos

    def truncate(self, pos=None):
        """Truncate stream
        """
        if pos is not None:
            self.seek(pos)
        self.seek_do()

        chunks, self.chunks = self.chunks[self.chunk_index + 1:], self.chunks[:self.chunk_index + 1]
        for chunk in chunks:
            self.store.Delete(chunk)
        self.chunk.truncate()
        self.chunk_dirty = True

    def flush(self):
        """Flush stream
        """
        if self.chunk_dirty:
            self.chunk_dirty = False
            chunk_desc = self.store.save(
                self.chunk.bytes() if not self.compress else
                zlib.compress(self.chunk.bytes(), self.compress))
            if self.chunk_index < len(self.chunks):
                self.chunks[self.chunk_index] = chunk_desc
            else:
                self.chunks.append(chunk_desc)

        header = json.dumps({
            'bufsize': self.bufsize,
            'chunks': self.chunks,
            'size': self.size,
            'compress': self.compress,
        }).encode()
        if self.header() != header:
            self.header(header)

    def close(self):
        return self.Flush()

    def dispose(self):
        self.flush()

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False


class Chunk(object):
    """Chunk

    Stream like object but with fixed capacity.
    """
    __slots__ = ('buf', 'cap', 'pos', 'size')

    def __init__(self, cap, data=None):
        self.buf = bytearray(cap)
        self.cap = cap
        self.pos = 0
        if data:
            self.size = len(data)
            self.buf[:self.size] = data
        else:
            self.size = 0

    def write(self, data):
        data_size = min(self.cap - self.pos, len(data))
        self.buf[self.pos:self.pos + data_size] = data[:data_size]
        self.pos += data_size
        self.size = max(self.size, self.pos)
        return data_size

    def read(self, size=None):
        data = (self.buf[self.pos:self.size] if size is None else
                self.buf[self.pos:min(self.pos + size, self.size)])
        self.pos += len(data)
        return bytes(data)

    def seek(self, pos, whence=0):
        if whence == 0:    # SEEK_SET
            self.pos = pos
        elif whence == 1:  # SEEK_CUR
            self.pos += pos
        elif whence == 2:  # SEEK_END
            self.pos = self.size + pos
        return self.pos

    def tell(self):
        return self.pos

    def truncate(self, pos=None):
        self.size = self.pos if pos is None else pos
        return self.size

    def bytes(self):
        return bytes(self.buf[:self.size])

    def __str__(self):
        return 'Chunk(data:{}, pos:{}, cap:{})'.format(self.bytes(), self.pos, self.cap)

    def __repr__(self):
        return str(self)



"""Micro virtual machine, compiler and abstract syntax tree for it
"""
import types
from ..monad import Cont, Result
from ..uniform import StringIO

__all__ = ('Expr', 'LoadArgExpr', 'LoadConstExpr', 'CallExpr',
           'GetAttrExpr', 'SetAttrExpr', 'GetItemExpr', 'SetItemExpr',
           'ReturnExpr', 'RaiseExpr', 'BindExpr', 'CmpExpr', 'IfExpr', 'WhileExpr',
           'Code', 'CodeGen')

HAS_ARG = 128

OP_POP = 1
OP_DUP = 2
OP_RETURN = 3
OP_RAISE = 4
OP_BIND = 5
OP_GETITEM = 6
OP_SETITEM = 7
OP_NOP = 8

OP_LDARG = 129    # argument index
OP_LDCONST = 130  # constant index
OP_CALL = 131     # number of arguments
OP_GETATTR = 132  # name as constant index
OP_SETATTR = 133  # name as constant index
OP_JMP = 134      # jump offset
OP_JMP_IF = 135   # jump offset
OP_JMP_IFN = 136  # jump offset
OP_COMPARE = 137  # compare operation

MAP_OP_NAME = {
    OP_POP:     'OP_POP',
    OP_DUP:     'OP_DUP',
    OP_RETURN:  'OP_RETURN',
    OP_RAISE:   'OP_RAISE',
    OP_BIND:    'OP_BIND',
    OP_GETITEM: 'OP_GETITEM',
    OP_SETITEM: 'OP_SETITEM',
    OP_NOP:     'OP_NOP',
    OP_LDARG:   'OP_LDARG',
    OP_LDCONST: 'OP_LDCONST',
    OP_CALL:    'OP_CALL',
    OP_GETATTR: 'OP_GETATTR',
    OP_SETATTR: 'OP_SETATTR',
    OP_JMP:     'OP_JMP',
    OP_JMP_IF:  'OP_JMP_IF',
    OP_JMP_IFN: 'OP_JMP_IFN',
    OP_COMPARE: 'OP_COMPARE',
}
MAP_NAME_OP = dict((name, op) for op, name in MAP_OP_NAME.items())

CMP_LE = 0   # <
CMP_LQ = 1   # <=
CMP_EQ = 2   # ==
CMP_NE = 3   # !=
CMP_GT = 4   # >
CMP_GE = 5   # >=
CMP_IN = 6   # in
CMP_NIN = 7  # not in
CMP_IS = 8   # is
CMP_ISN = 9  # is not
MAP_CMP_NAME = {
    CMP_LE:  '<',
    CMP_LQ:  '<=',
    CMP_EQ:  '==',
    CMP_NE:  '!=',
    CMP_GT:  '>',
    CMP_GE:  '>=',
    CMP_IN:  'in',
    CMP_NIN: 'not in',
    CMP_IS:  'is',
    CMP_ISN: 'is not'
}
MAP_NAME_CMP = dict((name, cmp) for cmp, name in MAP_CMP_NAME.items())


class Expr (object):
    __slots__ = tuple()

    def compile(self, gen):
        raise NotImplementedError()

    def code(self):
        gen = CodeGen()
        self.compile(gen)
        return gen.code()

    def __str__(self):
        return 'Nop'

    def __repr__(self):
        return str(self)


class LoadArgExpr(Expr):
    __slots__ = ('index',)

    def __init__(self, index):
        self.index = index

    def compile(self, gen):
        gen.emit(OP_LDARG, self.index)

    def __str__(self):
        return 'arg:{}'.format(self.index)


class LoadConstExpr(Expr):
    __slots__ = ('const',)

    def __init__(self, const):
        self.const = const

    def compile(self, gen):
        gen.emit(OP_LDCONST, gen.const(self.const))

    def __str__(self):
        if isinstance(self.const, (types.FunctionType, types.BuiltinFunctionType)):
            return getattr(self.const, '__qualname__', self.const.__name__)
        else:
            return repr(self.const)


class CallExpr(Expr):
    __slots__ = ('fn', 'args', 'kwargs',)

    def __init__(self, fn, *args, **kwargs):
        assert len(args) < 0xf or len(kwargs) > 0xf, "max argument count exceeded"
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def compile(self, gen):
        gen.load(self.fn)
        if self.args:
            for arg in self.args:
                gen.load(arg)
        if self.kwargs:
            for kw, arg in self.kwargs.items():
                gen.emit(OP_LDCONST, gen.const(kw))
                gen.load(arg)
        gen.emit(OP_CALL, len(self.args) | len(self.kwargs) << 4)

    def __str__(self):
        return ('{}({}{}{})'.format(self.fn,
                ', '.join(repr(arg) for arg in self.args),
                ', ' if self.args and self.kwargs else '',
                ', '.join('{}={}'.format(kw, repr(arg))
                          for kw, arg in self.kwargs.items())))


class GetAttrExpr(Expr):
    __slots__ = ('target', 'name',)

    def __init__(self, target, name):
        self.target = target
        self.name = name

    def compile(self, gen):
        gen.load(self.target)
        gen.emit(OP_GETATTR, gen.const(self.name))

    def __str__(self):
        return '{}.{}'.format(self.target, self.name)


class SetAttrExpr (Expr):
    __slots__ = ('target', 'name', 'value',)

    def __init__(self, target, name, value):
        self.target = target
        self.name = name
        self.value = value

    def compile(self, gen):
        gen.load(self.value)
        gen.load(self.target)
        gen.emit(OP_SETATTR, gen.const(self.name))

    def __str__(self):
        return '{}.{} = {}'.format(self.target, self.name, self.value)


class GetItemExpr (Expr):
    __slots__ = ('target', 'item',)

    def __init__(self, target, item):
        self.target = target
        self.item = item

    def compile(self, gen):
        gen.load(self.target)
        gen.load(self.item)
        gen.emit(OP_GETITEM)

    def __str__(self):
        return '{}[{}]'.format(self.target, self.item)


class SetItemExpr(Expr):
    __slots__ = ('target', 'item', 'value',)

    def __init__(self, target, item, value):
        self.target = target
        self.item = item
        self.value = value

    def compile(self, gen):
        gen.load(self.value)
        gen.load(self.target)
        gen.load(self.item)
        gen.emit(OP_SETITEM)

    def __str__(self):
        return '{}[{}] = {}'.format(self.target, self.item, self.value)


class ReturnExpr(Expr):
    __slots__ = ('result',)

    def __init__(self, result):
        self.result = result

    def compile(self, gen):
        gen.load(self.result)
        gen.emit(OP_RETURN)

    def __str__(self):
        return 'return {}'.format(self.result)


class RaiseExpr(Expr):
    __slots__ = ('error',)

    def __init__(self, error):
        self.error = error

    def compile(self, gen):
        gen.load(self.error)
        gen.emit(OP_RAISE)

    def __str__(self):
        return 'raise {}'.format(self.error)


class BindExpr(Expr):
    __slots__ = ('target',)

    def __init__(self, target):
        self.target = target

    def compile(self, gen):
        gen.load(self.target)
        gen.emit(OP_BIND)

    def __str__(self):
        return '<-{}'.format(self.target)


class CmpExpr(Expr):
    __slots__ = ('op', 'first', 'second',)

    def __init__(self, op, first, second):
        self.op = op if isinstance(op, int) else MAP_NAME_CMP[op]
        self.first = first
        self.second = second

    def compile(self, gen):
        gen.load(self.first)
        gen.load(self.second)
        gen.emit(OP_COMPARE, self.op)

    def __str__(self):
        return '{}{}{}'.format(self.first, self.op, self.second)


class IfExpr(Expr):
    __slots__ = ('cond', 'true', 'false',)

    def __init__(self, cond, true, false=None):
        self.cond = cond
        self.true = true
        self.false = false

    def compile(self, gen):
        jmp_label = gen.label()
        end_label = gen.label()

        # condition
        gen.load(self.cond)
        gen.emit(OP_JMP_IF, jmp_label)
        # false
        gen.load(self.false)
        gen.emit(OP_JMP, end_label)
        # true
        gen.label_mark(jmp_label)
        gen.load(self.true)
        # end
        gen.label_mark(end_label)

    def __str__(self):
        return '{} if {} else {}'.format(self.true, self.cond, self.false)


class WhileExpr(Expr):
    __slots__ = ('cond', 'body',)

    def __init__(self, cond, body):
        self.cond = cond
        self.body = body

    def compile(self, gen):
        begin_label = gen.label()
        end_label = gen.label()

        # condition
        gen.label_mark(begin_label)
        gen.load(self.cond)
        gen.emit(OP_JMP_IFN, end_label)
        # body
        gen.load(self.body)
        gen.emit(OP_POP)
        gen.emit(OP_JMP, begin_label)
        # end
        gen.label_mark(end_label)

    def __str__(self):
        return 'while {}: {}'.format(self.cond, self.body)


class CodeLabel(int):
    __slots__ = tuple()

    def __str__(self):
        return 'l:{}'.format(int.__str__(self))


class CodeGen(object):
    __slots__ = ('ops', 'consts', 'labels',)

    def __init__(self):
        self.ops = []
        self.consts = []
        self.labels = []

    def emit(self, op, arg=None):
        self.ops.append(op)
        if op & HAS_ARG:
            assert arg is not None, "this opcode must not have argument"
            self.ops.append(arg)

    def const(self, const):
        try:
            return self.consts.index(const)
        except Exception:
            self.consts.append(const)
            return len(self.consts) - 1

    def load(self, target):
        if isinstance(target, Expr):
            target.compile(self)
        else:
            self.emit(OP_LDCONST, self.const(target))

    def label(self):
        self.labels.append(None)
        return CodeLabel(len(self.labels) - 1)

    def label_mark(self, label):
        if self.labels[label] is not None:
            raise ValueError('label has already been marked')
        self.labels[label] = len(self.ops)

    def code(self):
        if None in self.labels:
            raise ValueError('not all labels has been marked')
        for index in range(len(self.ops)):
            if isinstance(self.ops[index], CodeLabel):
                self.ops[index] = self.labels[self.ops[index]]
        return Code(bytearray(self.ops),
                    tuple(self.consts) if self.consts else None)


class Code(tuple):
    __slots__ = tuple()

    def __new__(cls, ops, consts):
        return tuple.__new__(cls, (ops, consts))

    def __reduce__(self):
        return Code, tuple(self)

    @property
    def ops(self):
        return self[0]

    @property
    def consts(self):
        return self[1]

    def __call__(self, *args, **opts):
        monad = opts.get('monad', Cont)
        return monad.unit(None).bind(lambda _: self.eval(args, 0, [], monad))

    def eval(self, args, pos, stack, monad):
        ops, consts = self
        a_empty = tuple()
        try:
            ops_size = len(ops)
            while pos < ops_size:
                if ops[pos] & HAS_ARG:
                    op, arg = ops[pos:pos + 2]
                    pos = pos + 2
                else:
                    op, arg = ops[pos], '<not available>'
                    pos = pos + 1

                if op == OP_LDARG:
                    stack.append(args[arg])

                elif op == OP_LDCONST:
                    stack.append(consts[arg])

                elif op == OP_POP:
                    stack.pop()

                elif op == OP_DUP:
                    stack.append(stack[-1])

                elif op == OP_CALL:
                    if arg:
                        a_count, kw_count = arg & 0xf, arg >> 4
                        kw = {}
                        if kw_count:
                            for _ in range(kw_count):
                                val, key = stack.pop(), stack.pop()
                                kw[key] = val
                        a = a_empty
                        if a_count:
                            a, stack = stack[-a_count:], stack[:-a_count]
                        stack.append(stack.pop()(*a, **kw))
                    else:
                        stack.append(stack.pop()())

                elif op == OP_RETURN:
                    break

                elif op == OP_RAISE:
                    raise stack.pop()

                elif op == OP_BIND:
                    def eval_cont(val):
                        stack.append(val)
                        return self.eval(args, pos, list(stack), monad)
                    return monad.bind(stack.pop().__monad__(), eval_cont)

                elif op == OP_GETATTR:
                    stack.append(getattr(stack.pop(), consts[arg]))

                elif op == OP_SETATTR:
                    setattr(stack.pop(), consts[arg], stack.pop())

                elif op == OP_GETITEM:
                    item, target = stack.pop(), stack.pop()
                    stack.append(target[item])

                elif op == OP_SETITEM:
                    item, target, value = stack.pop(), stack.pop(), stack.pop()
                    target[item] = value

                elif op == OP_JMP:
                    pos = arg

                elif op == OP_JMP_IF:
                    if stack.pop():
                        pos = arg

                elif op == OP_JMP_IFN:
                    if not stack.pop():
                        pos = arg

                elif op == OP_COMPARE:
                    second, first = stack.pop(), stack.pop()
                    if arg == CMP_LE:
                        cmp = first < second
                    elif arg == CMP_LQ:
                        cmp = first <= second
                    elif arg == CMP_EQ:
                        cmp = first == second
                    elif arg == CMP_NE:
                        cmp = first != second
                    elif arg == CMP_GT:
                        cmp = first > second
                    elif arg == CMP_GE:
                        cmp = first >= second
                    elif arg == CMP_IN:
                        cmp = first in second
                    elif arg == CMP_NIN:
                        cmp == first not in second
                    elif arg == CMP_IS:
                        cmp = first is second
                    elif arg == CMP_ISN:
                        cmp = first is not second
                    else:
                        raise ValueError('unknown compare operation: {}'.format(arg))
                    stack.append(cmp)

                elif op == OP_NOP:
                    continue

                else:
                    raise ValueError('unknown operation code: {}'.format(op))

            return monad.unit(Result.from_value(stack.pop() if stack else None))

        except Exception:
            return monad.unit(Result.from_current_error())

    def __str__(self):
        ops, consts = self
        stream = StringIO()
        stream.write('Code(size:{}, consts:{}, opcodes:\n'.format(len(ops),
                     len(consts) if consts else 0))

        pos = 0
        while pos < len(ops):
            if ops[pos] & HAS_ARG:
                op, arg = ops[pos:pos + 2]
                stream.write('  {:>02} {:<10} '.format(pos, MAP_OP_NAME[op]))
                if op in (OP_LDCONST, OP_GETATTR, OP_SETATTR):
                    stream.write(repr(consts[arg]))
                elif op == OP_COMPARE:
                    stream.write(MAP_CMP_NAME[arg])
                else:
                    stream.write(repr(arg))
                stream.write('\n')
                pos = pos + 2

            else:
                op, arg = ops[pos], '<not available>'
                stream.write('  {:>02} {}\n'.format(pos, MAP_OP_NAME[op]))
                pos = pos + 1

        stream.seek(stream.tell() - 1)
        stream.write(')')
        return stream.getvalue()

    def __repr__(self):
        return str(self)



import io
import re
import struct
import unittest
import collections

from ..stream import Stream
from ..buffered import Buffer, BufferedStream
from ...monad import Result, async, do_return
from ...event import Event
from ...uniform import BrokenPipeError

__all__ = ('BufferTest', 'BufferedStreamTest',)


class BufferTest (unittest.TestCase):
    def test(self):
        buff = Buffer()

        buff.enqueue(b'01234')
        buff.enqueue(b'56789')
        buff.enqueue(b'01234')

        self.assertEqual(len(buff), 15)

        # single chunk
        self.assertEqual(buff.slice(3), b'012')
        self.assertEqual(tuple(buff.chunks), (b'01234', b'56789', b'01234',))

        # cross chunk
        self.assertEqual(buff.slice(6), b'012345')
        self.assertEqual(tuple(buff.chunks), (b'0123456789', b'01234',))

        # discard chunk
        buff.dequeue(3)
        self.assertEqual(len(buff), 12)
        self.assertEqual(buff.offset, 3)
        self.assertEqual(tuple(buff.chunks), (b'0123456789', b'01234',))

        # with offset
        self.assertEqual(buff.slice(3), b'345')
        self.assertEqual(tuple(buff.chunks), (b'0123456789', b'01234',))

        # discard cross chunk
        buff.dequeue(8)
        self.assertEqual(len(buff), 4)
        self.assertEqual(buff.offset, 1)
        self.assertEqual(tuple(buff.chunks), (b'01234',))

        buff.enqueue(b'56789')
        buff.enqueue(b'01234')

        # cross chunks with offset
        self.assertEqual(buff.slice(5), b'12345')
        self.assertEqual(tuple(buff.chunks), (b'0123456789', b'01234',))

        # peek all
        self.assertEqual(buff.slice(128), b'12345678901234')
        self.assertEqual(tuple(buff.chunks), (b'012345678901234',))

        buff.enqueue(b'56789')

        # discard all
        buff.dequeue(128)
        self.assertEqual(len(buff), 0)
        self.assertEqual(tuple(buff.chunks), tuple())

        for _ in range(3):
            buff.enqueue(b'0123456789')

        # discard with chunk cut
        buff.dequeue(6)
        self.assertEqual(len(buff), 24)
        self.assertEqual(buff.offset, 0)
        self.assertEqual(tuple(buff.chunks), (b'6789', b'0123456789', b'0123456789'))

        # discard edge
        buff.dequeue(14)
        self.assertEqual(len(buff), 10)
        self.assertEqual(buff.offset, 0)
        self.assertEqual(tuple(buff.chunks), (b'0123456789',))

        # discard less then half
        buff.dequeue(4)
        self.assertEqual(len(buff), 6)
        self.assertEqual(buff.offset, 4)
        self.assertEqual(tuple(buff.chunks), (b'0123456789',))

        # discard big
        buff.dequeue(128)
        self.assertEqual(len(buff), 0)
        self.assertEqual(buff.offset, 0)
        self.assertEqual(tuple(buff.chunks), tuple())


class BufferedStreamTest (unittest.TestCase):
    def test_read(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 8)

        stream.read(6)(res)
        stream.read_complete(b'012')
        self.assertEqual(res.pop(), b'012')

        stream.read(3)(res)
        stream.read_complete(b'012345')
        self.assertEqual(res.pop(), b'012')
        stream.read(6)(res)
        self.assertEqual(res.pop(), b'345')

        stream.read(3)(res)
        stream.read_complete(b'0123456789')
        self.assertEqual(res.pop(), b'012')
        stream.read(6)(res)
        self.assertEqual(res.pop(), b'34567')
        self.assertFalse(res)

    def test_read_until_size(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 8)

        stream.read_until_size(10)(res)
        self.assertFalse(res)
        stream.read_complete(b'0123456789')
        self.assertFalse(res)
        stream.read_complete(b'0123456789')
        self.assertEqual(res.pop(), b'0123456701')

        stream.read_until_size(4)(res)
        self.assertEqual(res.pop(), b'2345')
        self.assertFalse(res)

    def test_read_until_eof(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 1024)

        stream.read_until_eof()(res)
        stream.read_complete(b'01234')
        stream.read_complete(b'56789')
        self.assertFalse(res)
        stream.read_complete(Result.from_exception(BrokenPipeError()))
        self.assertEqual(res.pop(), b'0123456789')
        self.assertFalse(res)

    def test_read_until_sub(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 8)

        stream.read_until_sub(b';')(res)
        stream.read_complete(b'01234')
        self.assertFalse(res)
        stream.read_complete(b'56789;01')
        self.assertEqual(res.pop(), b'0123456789;')

        stream.read_until_sub(b';')(res)
        self.assertFalse(res)
        stream.read_complete(b'234;')
        self.assertEqual(res.pop(), b'01234;')
        self.assertFalse(res)

    def test_read_until_regex(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 1024)
        regex = re.compile(br'([^=]+)=([^&]+)&')

        stream.read_until_regex(regex)(res)
        stream.read_complete(b'key_0=')
        self.assertFalse(res)
        stream.read_complete(b'value_0&key_1')
        self.assertEqual(res.pop()[0], b'key_0=value_0&')

        stream.read_until_regex(regex)(res)
        self.assertFalse(res)
        stream.read_complete(b'=value_1&tail')
        self.assertEqual(res.pop()[0], b'key_1=value_1&')

        stream.read(4)(res)
        self.assertEqual(res.pop(), b'tail')
        self.assertFalse(res)

    def test_write(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 8)

        stream.write(b'0123456')(res)
        stream.write_complete(10)
        self.assertEqual(res.pop(), 7)
        self.assertEqual(stream.written, b'')

        stream.write(b'7')(res)  # flusher started
        self.assertEqual(res.pop(), 1)
        stream.write(b'89')(res)
        self.assertEqual(res.pop(), 2)
        stream.write_complete(10)
        self.assertEqual(stream.written, b'01234567')

        stream.write_complete(2)
        self.assertEqual(stream.written, b'0123456789')
        self.assertFalse(res)

        stream.written_stream.truncate(0)
        stream.written_stream.seek(0)
        stream.write(b'X' * 17)(res)  # blocked
        self.assertFalse(res)
        stream.write_complete(8)
        self.assertFalse(res)
        stream.write_complete(8)
        self.assertFalse(res)
        stream.write_complete(1)
        self.assertEqual(stream.written, b'X' * 17)
        self.assertEqual(res.pop(), 17)
        self.assertFalse(res)

    def test_bytes(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 1024)
        bytes = b'some bytes string'

        stream.write_bytes(bytes)
        stream.flush()(res)
        self.assertFalse(res)
        stream.write_complete(1024)
        self.assertEqual(res.pop(), None)

        stream.read_bytes()(res)
        stream.read_complete(stream.written)
        self.assertEqual(res.pop(), bytes)
        self.assertFalse(res)

    def test_struct_list(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 1024)

        struct_type = struct.Struct('>H')
        struct_list = [23, 16, 10, 32, 45, 18]

        stream.write_struct_list(struct_list, struct_type)
        stream.flush()()
        stream.write_complete(1024)

        stream.read_struct_list(struct_type)(res)
        stream.read_complete(stream.written)
        self.assertEqual(res.pop(), struct_list)
        self.assertFalse(res)

    def test_struct_list_complex(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 1024)

        struct_type = struct.Struct('>HH')
        struct_list = [(23, 0), (16, 1), (10, 2), (32, 3), (45, 4), (18, 5)]

        stream.write_struct_list(struct_list, struct_type, True)
        stream.flush()()
        stream.write_complete(1024)

        stream.read_struct_list(struct_type, True)(res)
        stream.read_complete(stream.written)
        self.assertEqual(res.pop(), struct_list)
        self.assertFalse(res)

    def test_bytes_list(self):
        res = ResultQueue()
        stream = BufferedStream(DummyStream(), 1024)
        bytes_list = [b'one', b'two', b'three', b'four', b'five']

        stream.write_bytes_list(bytes_list)
        stream.flush()()
        stream.write_complete(1024)

        stream.read_bytes_list()(res)
        stream.read_complete(stream.written)
        self.assertEqual(res.pop(), bytes_list)
        self.assertFalse(res)


class DummyStream(Stream):
    def __init__(self):
        Stream.__init__(self)
        self.initing()
        self.written_stream = io.BytesIO()
        self.read_complete = Event()
        self.write_complete = Event()

    @async
    def read(self, size):
        with self.reading:
            do_return((yield self.read_complete)[:size])

    @async
    def write(self, data):
        with self.writing:
            size = yield self.write_complete
            self.written_stream.write(data[:size])
            do_return(min(size, len(data)))

    @property
    def written(self):
        return self.written_stream.getvalue()


class ResultQueue(object):
    def __init__(self):
        self.queue = collections.deque()

    def __call__(self, val):
        self.queue.append(val)

    def pop(self):
        return self.queue.popleft().value

    def __len__(self):
        return len(self.queue)

    def __bool__(self):
        return bool(self.queue)
    __nonzero__ = __bool__



import unittest
import collections
from ..hub import Hub, Address, Sender, pair

__all__ = ('HubTest',)


class HubTest(unittest.TestCase):
    def test_addr(self):
        addr = Hub.local().addr()
        self.assertEqual(len(addr), 1)

        addr_111 = addr.route((111,))
        self.assertEqual(addr_111, Address((111,)))

        self.assertEqual(addr_111.unroute(), addr)

    def test_sender_receiver(self):
        res = ResultQueue()
        recv, send = pair()

        with self.assertRaises(ValueError):
            send.send('no-recv')  # no receiver

        # send
        recv(res)
        with self.assertRaises(ValueError):
            recv(lambda *a: res(*a))  # multiple handlers
        send.send('1')
        with self.assertRaises(ValueError):
            send.send('no-recv')  # no receiver
        self.assertEqual(res.pop(), ('1', recv.addr, None))
        self.assertFalse(res)

        # send with source
        recv(res)
        send.send('2', 'sender')
        self.assertEqual(res.pop(), ('2', recv.addr, 'sender'))

        # call
        recv(res)
        send('3')(res)
        self.assertEqual(len(Hub.local()), 1)  # handler will be cleaned up
        msg, dst, src = res.pop()
        self.assertEqual(msg, '3')
        self.assertEqual(dst, recv.addr)
        self.assertTrue(isinstance(src, Sender))
        self.assertFalse(res)

        with self.assertRaises(ValueError):
            send('4')(res)  # send on address without a receiver
            res.pop()[0].value

        src.send('5')
        self.assertEqual(res.pop()[0].value, '5')
        self.assertFalse(res)

        self.assertEqual(len(Hub.local()), 0)

    def test_faulty_handler(self):
        recv, send = pair()

        def faulty(msg, dst, src):
            if msg == '0':
                raise RuntimeError()
            return False
        recv(faulty)

        with self.assertRaises(RuntimeError):
            send.send('0')
        self.assertEqual(len(Hub.local()), 1)
        send.send('1')
        self.assertEqual(len(Hub.local()), 0)

    def test_reentrancy(self):
        res = ResultQueue()
        recv, send = pair()

        def handler(msg, dst, src):
            res(msg)
            if msg == 'first':
                send.send('second')
                return True
            else:
                return False
        recv(handler)

        send.send('first')
        self.assertEqual(res.pop(), ('first',))
        self.assertEqual(res.pop(), ('second',))


class ResultQueue(object):
    def __init__(self):
        self.queue = collections.deque()

    def __call__(self, *args):
        self.queue.append(args)
        return False

    def pop(self):
        return self.queue.popleft()

    def __len__(self):
        return len(self.queue)

    def __bool__(self):
        return bool(self.queue)
    __nonzero__ = __bool__



"""Asynchronous application framework
"""
import os
import sys
import select

__all__ = ('PRETZEL_RECLIMIT', 'PRETZEL_BUFSIZE', 'PRETZEL_TEST_TIMEOUT',
           'PRETZEL_POLLER',)

# Environment configurable pretzel variables
PRETZEL_RECLIMIT = int(os.environ.get('PRETZEL_RECLIMIT', '8192'))
PRETZEL_BUFSIZE = int(os.environ.get('PRETZEL_BUFSIZE', '65536'))
PRETZEL_TEST_TIMEOUT = int(os.environ.get('PRETZEL_TEST_TIMEOUT', '5'))
PRETZEL_POLLER = os.environ.get('PRETZEL_POLLER',
                                'epoll' if hasattr(select, 'epoll') else
                                'kqueue' if hasattr(select, 'kqueue') else
                                'select')


# Increase of recursion limit is desirable as in case of long sequence of
# instantly resolved monads do_block may exhaust it.
sys.setrecursionlimit(int(os.environ.get('PRETZEL_RECLIMIT', '8192')))


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import tests, monad, core, stream, remoting, store

    suite = TestSuite()
    for test in (tests, monad, core, stream, remoting, store):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite


def load_bench(runner):
    """Load benchmarks protocol
    """
    from . import remoting

    for module in (remoting,):
        runner.add_module(module)



"""Connection
"""
import io
import sys
import textwrap
from pickle import Pickler, Unpickler
from ..hub import Hub, Sender, pair
from ..proxy import Proxy
from ..expr import LoadConstExpr, LoadArgExpr
from ...uniform import reraise
from ...event import Event
from ...core import Core
from ...monad import Result, async, do_return
from ...state_machine import StateMachine
from ...dispose import CompDisp

__all__ = ('Connection',)


class Connection(object):
    """Connection
    """
    STATE_INIT = 0
    STATE_CONNI = 1
    STATE_CONND = 2
    STATE_DISP = 3
    STATE_GRAPH = StateMachine.compile_graph({
        STATE_INIT:  (STATE_CONNI, STATE_DISP),
        STATE_CONNI: (STATE_CONND, STATE_DISP),
        STATE_CONND: (STATE_DISP,),
        STATE_DISP:  (STATE_DISP,)
    })
    STATE_NAMES = ('not-connected', 'connecting', 'connected', 'disposed',)

    def __init__(self, hub=None, core=None):
        self.hub = hub or Hub.local()
        self.core = core or Core.local()
        self.flags = {}
        self.module_map = {}

        self.receiver, self.sender = pair(hub=self.hub)
        self.state = StateMachine(self.STATE_GRAPH, self.STATE_NAMES)
        self.recv_ev = Event()

        ## dispose
        self.dispose = CompDisp()
        self.dispose.add_action(lambda: self.do_disconnect())
        self.dispose.add(self.receiver)
        self.dispose.add_action(lambda: self.state(self.STATE_DISP))

        ## marshaling
        PACK_ROUTE = 1
        PACK_UNROUTE = 2

        class pickler_type(Pickler):
            def persistent_id(this, target):
                if isinstance(target, Sender):
                    if target.hub is not self.hub:
                        raise ValueError('sender\'s hub must match hub used by connection')
                    if target.addr == self.sender.addr:
                        # This sender was previously received from this connection
                        # so it must not be routed again.
                        return PACK_UNROUTE, target.addr.unroute()
                    else:
                        # Sender must be routed
                        return PACK_ROUTE, target.addr
                return None

        self.pickler_type = pickler_type

        class unpickler_type(Unpickler):
            def persistent_load(this, state):
                pack, args = state
                if pack == PACK_ROUTE:
                    return Sender(self.hub, args.route(self.sender.addr))
                elif pack == PACK_UNROUTE:
                    return Sender(self.hub, args if args else self.sender.addr)
                else:
                    raise ValueError('Unknown pack type: {}'.format(pack))

            def find_class(this, modname, name):
                modname = self.module_map.get(modname, modname)
                module = sys.modules.get(modname, None)
                if module is None:
                    __import__(modname)
                    module = sys.modules[modname]
                if getattr(module, '__initializing__', False):
                    # Module is being imported. Interrupt unpickling.
                    raise InterruptError()
                return getattr(module, name)

        self.unpickler_type = unpickler_type

    def __call__(self, target):
        """Create proxy object from provided pickle-able constant.
        """
        return Proxy(self.sender, LoadConstExpr(target))

    @property
    def connected(self):
        return self.state.state == self.STATE_CONND

    @async
    def connect(self, target=None):
        self.state(self.STATE_CONNI)
        try:
            def send(msg, dst, src):
                stream = io.BytesIO()
                self.pickler_type(stream, -1).dump((msg, dst, src))
                self.do_send(stream.getvalue())
                return True
            self.receiver(send)
            yield self.do_connect(target)
            self.state(self.STATE_CONND)
        except Exception:
            error = sys.exc_info()
            self.dispose()
            reraise(*error)
        do_return(self)

    @async
    def do_connect(self, target):
        """Connect implementation
        """

    def do_disconnect(self):
        """Disconnect implementation
        """

    def do_send(self, msg):
        """Send packed message to other peer
        """
        raise NotImplementedError()

    @async
    def do_recv(self, msg):
        """Handle remote packed message
        """
        # Detachment from  current coroutine is vital here because if handler
        # tries to create nested core loop to resolve future synchronously
        # (i.g. importer proxy) it can block dispatching coroutine.
        yield self.core.schedule()
        self.recv_ev(msg)

        while True:
            src = None
            try:
                msg, dst, src = self.unpickler_type(io.BytesIO(msg)).load()
                dst = dst.unroute()  # strip remote connection address

                if dst:
                    # After striping remote connection address, destination
                    # is not empty so it needs to be routed.
                    self.hub.send(msg, dst, src)
                else:
                    if msg is None:  # pragma: no cover (covered by remote path)
                        self.dispose()
                        return
                    if src is None:
                        yield msg(self)
                    else:
                        src.send((yield msg(self)))
                break
            except InterruptError:  # pragma: no cover (covered by remote path)
                # Required module is being imported right now. Wait for pending
                # imports. We also need to postpone dispatch of the message
                # because, message which resumed this method, may in the same
                # time complete pending import.
                yield self.recv_ev
                yield self.core.schedule()
            except Exception:  # pragma: no cover (covered by remote path)
                if not self.disposed:
                    err = Result.from_current_error()
                    if src is not None:
                        src.send(err)
                    else:
                        err.trace(banner=lambda: textwrap.dedent("""\
                            [connection] impossible to send error response to message:
                              {} -> {}""").format(msg, dst))
                break

    def __monad__(self):
        return self.connect()

    def __reduce__(self):
        return ConnectionProxy, (self.sender, LoadArgExpr(0), self.flags)

    @property
    def disposed(self):
        return self.state.state == self.STATE_DISP

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        flags = ''.join('{}:{}, '.format(key, val) for key, val in self.flags.items())
        return ('{}({}state:{}, addr:{})'.format(type(self).__name__,
                flags, self.state.state_name(), self.sender.addr))

    def __repr__(self):
        return str(self)


class ConnectionProxy(Proxy):
    """Connection proxy object
    """
    __slots__ = Proxy.__slots__ + ('_flags',)

    def __init__(self, sender, expr, flags):
        Proxy.__init__(self, sender, expr)
        self._flags = flags

    def __call__(self, target):
        return Proxy(self._sender, LoadConstExpr(target))

    def __reduce__(self):
        return ConnectionProxy, (self._sender, self._expr, self._flags)

    def __str__(self):
        flags = ''.join('{}:{}, '.format(key, val) for key, val in self._flags.items())
        return '{}({}addr:{})'.format(type(self).__name__, flags,
                                      self._sender.addr if self._sender else None)


class InterruptError (BaseException):
    """Interrupt helper exception type
    """



"""Send-able lazy proxy object
"""
from .hub import pair
from .expr import (LoadArgExpr, LoadConstExpr, CallExpr, GetAttrExpr,
                   GetItemExpr, BindExpr)
from ..monad import async, do_return

__all__ = ('Proxy', 'proxify', 'proxify_func',)


class Proxy(object):
    """Send-able lazy proxy object
    """
    __slots__ = ('_sender', '_expr', '_code',)

    def __init__(self, sender, expr):
        self._sender = sender
        self._expr = expr
        self._code = None

    def __call__(self, *args, **kwargs):
        return Proxy(self._sender, CallExpr(self._expr, *args, **kwargs))

    def __getattr__(self, name):
        return Proxy(self._sender, GetAttrExpr(self._expr, name))

    def __getitem__(self, item):
        return Proxy(self._sender, GetItemExpr(self._expr, item))

    def __invert__(self):
        return Proxy(self._sender, BindExpr(self._expr))

    def __rshift__(self, func):
        return Proxy(self._sender, CallExpr(LoadConstExpr(func), self._expr))

    def __monad__(self):
        if self._code is None:
            self._code = self._expr.code()
        return self._sender(self._code)

    def __reduce__(self):
        return type(self), (self._sender, self._expr)

    def __str__(self):
        return ('{}(addr:{}, expr:{})'.format(type(self).__name__,
                self._sender.addr if self._sender else None, self._expr))

    def __repr__(self):
        return str(self)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        sender, self._sender = self._sender, None
        if sender is not None:
            sender.try_send(None)
        return False


def proxify(target, dispose=None, hub=None):
    """Create send-able proxy object from target object
    """
    if isinstance(target, Proxy):
        return target >> proxify

    def proxy_handler(code, dst, src):
        if code is None:
            if dispose is None or dispose:
                getattr(target, '__exit__', lambda *_: None)(None, None, None)
            return False  # unsubscribe proxy handler
        code(target)(lambda val: val.trace() if src is None else src.send(val))
        return True

    recv, send = pair(hub=hub)
    recv(proxy_handler)
    return Proxy(send, LoadArgExpr(0))


class FuncProxy(object):
    """Function object proxy
    """
    __slots__ = ('sender',)

    def __init__(self, sender):
        self.sender = sender

    def __call__(self, *args, **kwargs):
        if self.sender is None:
            raise RuntimeError('function proxy has been disposed')
        return self.sender((args, kwargs))

    def dispose(self):
        sender, self.sender = self.sender, None
        if sender is not None:
            sender.try_send(None)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return '{}(addr:{})'.format(type(self).__name__,
                                    self.sender.addr if self.sender else None)

    def __repr__(self):
        return str(self)


def proxify_func(func, hub=None):
    """Proxify asynchronous function
    """
    @async
    def func_caller(msg):
        args, kwargs = msg
        do_return((yield func(*args, **kwargs)))

    def func_handler(msg, dst, src):
        if msg is None:
            return False
        else:
            func_caller(msg)(lambda val: val.trace() if src is None else src.send(val))
            return True

    recv, send = pair(hub=hub)
    recv(func_handler)
    return FuncProxy(send)



"""Base asynchronous stream type
"""
from collections import defaultdict
from .. import PRETZEL_BUFSIZE
from ..monad import async
from ..state_machine import StateMachine
from ..uniform import BrokenPipeError

__all__ = ('Stream',)


class Stream(object):
    """Base class for asynchronous streams
    """
    FLAG_READ = 0b00001
    FLAG_WRITE = 0b00010
    FLAG_IDLE = 0b00100

    STATE_NONE = 0b00000
    STATE_READ = FLAG_READ | FLAG_IDLE  # 0b00101
    STATE_WRITE = FLAG_WRITE | FLAG_IDLE  # 0b00110
    STATE_READ_WRITE = FLAG_READ | FLAG_WRITE | FLAG_IDLE  # 0b00111
    STATE_IDLE = FLAG_IDLE  # 0b00100
    STATE_INIT = 0b01000
    STATE_DISPOSED = 0b10000
    STATE_GRAPH = StateMachine.compile_graph({
        STATE_NONE: (STATE_INIT, STATE_DISPOSED),
        STATE_READ: (STATE_IDLE, STATE_READ_WRITE, STATE_DISPOSED),
        STATE_WRITE: (STATE_IDLE, STATE_READ_WRITE, STATE_DISPOSED),
        STATE_READ_WRITE: (STATE_READ, STATE_WRITE, STATE_DISPOSED),
        STATE_INIT: (STATE_NONE, STATE_IDLE, STATE_DISPOSED),
        STATE_IDLE: (STATE_READ, STATE_WRITE, STATE_DISPOSED),
        STATE_DISPOSED: (STATE_DISPOSED,),
    })
    STATE_NAMES = defaultdict(lambda: 'invalid')
    STATE_NAMES.update({
        STATE_NONE: 'not-inited',
        STATE_READ: 'read',
        STATE_WRITE: 'write',
        STATE_READ_WRITE: 'read-write',
        STATE_INIT: 'initing',
        STATE_IDLE: 'idle',
        STATE_DISPOSED: 'disposed'
    })

    def __init__(self):
        self.state = StateMachine(self.STATE_GRAPH, self.STATE_NAMES)
        self.reading = StateFlagScope(self.state, self.FLAG_READ)
        self.writing = StateFlagScope(self.state, self.FLAG_WRITE)
        self.initing = StateTransScope(self.state, self.STATE_INIT,
                                       self.STATE_IDLE, self.STATE_NONE)

    def fileno(self):
        raise NotImplementedError()

    @async
    def read(self, size):
        """Read data

        Size of returned data expected to be in range [1..size], otherwise
        BrokenPipeError must be raised. Reading scope must be entered for
        duration of reading.
        """
        with self.reading:
            raise NotImplementedError()

    @async
    def write(self, data):
        """Write data

        Returns length of written data. Writing scope must be entered for
        duration of writing.
        """
        with self.writing:
            raise NotImplementedError()

    @async
    def flush(self):
        """Flush write buffers

        Must hook continuation to the current flush if any or start new one.
        """

    @async
    def flush_and_dispose(self):
        """Flush data and dispose stream
        """
        yield self.flush()
        self.dispose()

    @async
    def copy_to(self, stream, bufsize=None):
        """Copy content of this stream to different stream

        Stream can be either asynchronous stream or python binary stream.
        """
        bufsize = bufsize or PRETZEL_BUFSIZE
        if isinstance(stream.write(b''), int):
            # destination stream is synchronous python stream
            try:
                while True:
                    stream.write((yield self.read(bufsize)))
            except BrokenPipeError:
                pass
            stream.flush()
        else:
            try:
                while True:
                    yield stream.write((yield self.read(bufsize)))
            except BrokenPipeError:
                pass
            yield stream.flush()

    @property
    def disposed(self):
        return self.state.state == self.STATE_DISPOSED

    def dispose(self):
        return self.state(self.STATE_DISPOSED)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return '{}(state:{})'.format(type(self).__name__, self.state.state_name())

    def __repr__(self):
        return str(self)


class StateFlagScope(object):
    __slots__ = ('state', 'flag')

    def __init__(self, state, flag):
        self.state = state
        self.flag = flag

    def __enter__(self):
        self.state(self.state.state | self.flag)
        return self

    def __exit__(self, et, eo, tb):
        self.state(self.state.state & ~self.flag)
        return False

    def __bool__(self):
        return bool(self.state.state & self.flag)
    __nonzero__ = __bool__

    def __str__(self):
        return str(bool(self))
    __repr__ = __str__


class StateTransScope(object):
    __slots__ = ('state', 'prog', 'succ', 'fail',)

    def __init__(self, state, prog, succ, fail):
        self.state = state
        self.prog = prog
        self.succ = succ
        self.fail = fail

    def __call__(self):
        self.state(self.prog)
        self.state(self.succ)

    def __enter__(self):
        self.state(self.prog)
        return self

    def __exit__(self, et, eo, tb):
        self.state(self.succ if et is None or not issubclass(et, Exception) else
                   self.fail)
        return False

    def __bool__(self):
        return bool(self.state.state == self.prog)
    __nonzero__ = __bool__

    def __str__(self):
        return str(bool(self))
    __repr__ = __str__



"""Helper classes to compose disposable objects
"""
from .monad import async_block

__all__ = ('Disposable', 'FuncDisp', 'CompDisp',)


class Disposable(object):
    """Base disposable
    """
    __slots__ = tuple()

    def __call__(self):
        return self.dispose()

    @property
    def disposed(self):
        raise NotImplementedError()

    def __bool__(self):
        return self.disposed

    def __nonzero__(self):
        return self.disposed

    def dispose(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return '{}(disposed:{})'.format(type(self).__name__, self.disposed)

    def __repr__(self):
        return str(self)


class FuncDisp(Disposable):
    """Action based disposable object
    """
    __slots__ = ('action',)

    def __init__(self, action=None):
        self.action = action

    @property
    def disposed(self):
        return self.action is None

    def dispose(self):
        action, self.action = self.action, None
        if action is None:
            return False
        else:
            action()
            return True


class CompDisp(Disposable):
    """Composite disposable

    Treat multiple disposables as one.
    """
    __slots__ = ('disps',)

    def __init__(self, disps=None):
        self.disps = []
        if disps:
            try:
                for disp in disps:
                    self.add(disp)
            except Exception:
                self.dispose()
                raise

    def add(self, disp):
        disp.__enter__()
        if self.disps is None:
            disp.__exit__(None, None, None)
        else:
            self.disps.append(disp)
        return disp

    def add_action(self, action):
        return self.add(FuncDisp(action))

    def __iadd__(self, disp):
        self.add(disp)
        return self

    def remove(self, disp):
        try:
            if self.disps is None:
                return False
            self.disps.remove(disp)
            return True
        except ValueError:
            return False
        finally:
            disp.__exit__(None, None, None)

    def __isub__(self, disp):
        self.remove(disp)
        return self

    def __len__(self):
        return len(self.disps) if self.disps else 0

    @property
    def disposed(self):
        return self.disps is None

    def dispose(self):
        disps, self.disps = self.disps, None
        if disps is None:
            return False
        else:
            for disp in reversed(disps):
                disp.__exit__(None, None, None)
            return True

    def __monad__(self):
        """Wait for composite disposable object to be disposed
        """
        return async_block(lambda ret: self.add_action(lambda: ret(None)))



import io
import random
import unittest
from ..store import StreamStore

__all__ = ('StoreTest',)


class StoreTest(unittest.TestCase):
    """Store unit tests
    """
    def test_simple(self):
        """Simple tests
        """
        stream = io.BytesIO()
        with StreamStore(stream, 1) as store:
            # empty
            self.assertEqual(store.save(b''), 0)
            self.assertEqual(store.load(0), b'')

            # unnamed
            desc_data = b'some test data'
            desc = store.save(desc_data)
            self.assertEqual(store.load(desc), desc_data)

            # named
            name, name_data = b'name', b'some test value'
            store[name] = name_data
            self.assertEqual(store[name], name_data)

        with StreamStore(stream, 1) as store:
            self.assertEqual(store.load(desc), desc_data)
            self.assertEqual(store[name], name_data)

    def test_stress(self):
        """Stress tests
        """
        count = 1 << 14
        datas = [str(random.randint(0, count)).encode() * random.randint(1, 16)
                 for _ in range(count)]

        stream = io.BytesIO()
        descs = []

        with StreamStore(stream) as store:
            for data in datas:
                descs.append(store.save(data))

        with StreamStore(stream) as store:
            for data, desc in zip(datas, descs):
                self.assertEqual(data, store.load(desc))

            # delete half
            for desc in descs[int(count / 2):]:
                store.delete(desc)

            datas = datas[:int(count / 2)]
            descs = descs[:int(count / 2)]

        with StreamStore(stream) as store:
            for data, desc in zip(datas, descs):
                self.assertEqual(data, store.load(desc))

            # create
            for _ in range(count):
                data = str(random.randint(0, count)).encode() * random.randint(1, 16)
                datas.append(data)
                descs.append(store.save(data))

        with StreamStore(stream) as store:
            for data, desc in zip(datas, descs):
                self.assertEqual(data, store.load(desc))



"""Utility functions and types
"""
from functools import wraps

__all__ = ('lazy',)


def lazy(func, *args, **kwargs):
    """Lazy value
    """
    @wraps(func)
    def lazy_func():
        if not value:
            value.append(func(*args, **kwargs))
        return value[0]
    value = []
    return lazy_func



"""Asynchronous utility functions

Function to work and create continuation monads with embedded value of
result monad type.
"""
from functools import wraps
from collections import deque
from .do import do
from .do_green import do_green
from .cont import Cont
from .result import Result, callsite_banner
from ..event import Event

__all__ = ('async', 'async_green', 'async_block', 'async_any', 'async_all',
           'async_limit', 'async_single',)


def async(block):
    """Better "do" block for continuation monad

    It is also possible to run returned continuation multiple times, which
    is not possible with "do" block.
    """
    do_block = do(Cont)(block)
    return wraps(block)(lambda *a, **kw: Cont(
                        lambda ret: do_block(*a, **kw).run(ret)))


def async_green(block):
    """Better "do_green" block for continuation monad

    It is also possible to run returned continuation multiple times, which
    is not possible with "do_green" block.
    """
    do_block = do_green(Cont)(block)
    return wraps(block)(lambda *a, **kw: Cont(
                        lambda ret: do_block(*a, **kw).run(ret)))


def async_block(block):
    """Create continuation from block

    Behaves similar to callcc but returned continuation is not resolved when
    block is left. And if block raises and error returned continuation will be
    resolved with result monad containing this error.
    """
    def async_block(ret):
        def block_ret(val=None):
            try:
                if block_done[0]:  # pragma: no cover
                    raise RuntimeError('return handler has been called twice')
                block_done[0] = True
                ret(val if isinstance(val, Result) else Result.from_value(val))
            except Exception:  # pragma: no cover
                Result.from_current_error().trace(banner=banner)
        try:
            block_done = [False]
            block(block_ret)
        except Exception:
            if block_done[0]:
                Result.from_current_error().trace(banner=banner)
            else:
                block_ret(Result.from_current_error())
    banner = callsite_banner('[async_block] return function failed')
    return Cont(async_block)


def async_any(conts):
    """Any continuation

    Resolved with the result of first continuation to be resolved.
    """
    conts = tuple(cont.__monad__() for cont in conts)
    if not conts:
        raise ValueError('continuation set is empty')

    @async_block
    def any_cont(ret):
        # Context structure: [completed?]
        ctx = [False]

        def cont_ret(val):
            if not ctx[0]:
                ctx[0] = True
                ret(val)

        for cont in conts:
            cont.run(cont_ret)
    return any_cont


def async_all(conts):
    """All continuation

    Resolved with the list of results of all continuations.
    """
    conts = tuple(cont.__monad__() for cont in conts)
    if not conts:
        return Cont.unit(Result.unit(tuple()))

    @async_block
    def all_cont(ret):
        # Context structure: [val_0 .. val_N, count]
        ctx = [None] * (len(conts) + 1)
        ctx[-1] = len(conts)

        def cont_register(index, cont):
            def cont_ret(val):
                ctx[index] = (val if isinstance(val, Result) else Result.from_value(val))
                ctx[-1] -= 1
                if not ctx[-1]:
                    ret(Result.sequence(ctx[:-1]))
            cont.run(cont_ret)

        for index, cont in enumerate(conts):
            cont_register(index, cont)
    return all_cont


def async_limit(limit):
    """Limit number of simultaneously executing coroutines

    Returns decorator, which limit number of unresolved continuations to
    specified limit value.
    """
    def async_limit(func):
        worker_count = [0]
        worker_queue = deque()

        @async
        def worker():
            try:
                worker_count[0] += 1
                while worker_queue:
                    ret, args, kwargs = worker_queue.popleft()
                    try:
                        ret((yield func(*args, **kwargs)))
                    except Exception:
                        ret(Result.from_current_error())
            finally:
                worker_count[0] -= 1

        @wraps(func)
        def func_limit(*args, **kwargs):
            @async_block
            def cont(ret):
                worker_queue.append((ret, args, kwargs))
                if worker_count[0] < limit:
                    worker()()
            return cont

        return func_limit
    return async_limit


def async_single(func, *args, **kwargs):
    """Singleton asynchronous action

    If there is current non finished continuation, all calls would return this
    continuation, otherwise new continuation will be started.
    """
    @wraps(func)
    def func_single():
        @async_block
        def cont(ret):
            done.on_once(ret)
            if len(done.handlers) == 1:
                func(*args, **kwargs).__monad__()(done)
        return cont
    done = Event()
    return func_single



import unittest
import operator
import itertools
from ..dispose import FuncDisp, CompDisp
from ..monad import async

__all__ = ('FuncDispTest', 'CompDispTest',)


class FuncDispTest(unittest.TestCase):
    def test(self):
        def disp():
            res[0] += 1
        res = [0]

        with FuncDisp(disp) as d0:
            self.assertFalse(d0)
            self.assertEqual(res[0], 0)
        self.assertTrue(d0)
        self.assertEqual(res[0], 1)
        d0.dispose()
        self.assertEqual(res[0], 1)

        d1 = FuncDisp(disp)
        self.assertEqual(res[0], 1)
        d1.dispose()
        self.assertEqual(res[0], 2)


class CompDispTest(unittest.TestCase):
    def test(self):
        ctx = [0, 0, 0, 0]
        d0, d1, d2, d3 = map(FuncDisp, (
            lambda: operator.setitem(ctx, 0, 1),
            lambda: operator.setitem(ctx, 1, 1),
            lambda: operator.setitem(ctx, 2, 1),
            lambda: operator.setitem(ctx, 3, 1)))

        d = CompDisp((d0,))
        self.assertFalse(d)
        self.assertEqual(ctx, [0, 0, 0, 0])
        d += d1
        with d:
            self.assertEqual(ctx, [0, 0, 0, 0])
        self.assertTrue(d)
        self.assertEqual(ctx, [1, 1, 0, 0])
        d += FuncDisp(lambda: d.add(d3))
        d.dispose()
        self.assertEqual(ctx, [1, 1, 0, 1])

    def test_order(self):
        ctx = [0, 0, 0, 0]
        order = itertools.count()
        d0, d1, d2, d3 = map(FuncDisp, (
            lambda: operator.setitem(ctx, 0, next(order)),
            lambda: operator.setitem(ctx, 1, next(order)),
            lambda: operator.setitem(ctx, 2, next(order)),
            lambda: operator.setitem(ctx, 3, next(order))))
        next(order)
        with CompDisp() as d:
            d.add(d0)
            d.add(d2)
            d.add(d1)
            d.add(d3)
            self.assertEqual(ctx, [0, 0, 0, 0])
        self.assertEqual(ctx, [4, 2, 3, 1])

    def test_async(self):
        with CompDisp() as d:
            future = d.__monad__().future()
            self.assertFalse(future.completed)
        self.assertTrue(future.completed)
        self.assertTrue(d.__monad__().future().completed)



"""Store based B+Tree provider
"""
import io
import sys
import json
import zlib
import struct
import codecs
import binascii
import operator
import functools
from bisect import bisect
if sys.version_info[0] > 2:
    import pickle
else:
    import cPickle as pickle
from .provider import BPTreeProvider
from ..bptree import BPTreeNode, BPTreeLeaf
from ...alloc import StoreBlock
from ...serialize import Serializer

__all__ = ('StoreBPTreeProvider',)


class StoreBPTreeProvider(BPTreeProvider):
    """Store based B+Tree provider

    Keeps serialized(possible compressed) nodes inside store. Keys and values
    are serialized according to specified type. Possible values for type are
    'bytes', 'pickle:protocol', 'struct:struct_type', 'json'.
    """

    order_default = 128
    type_default = 'pickle:{}'.format(pickle.HIGHEST_PROTOCOL)
    compress_default = 9
    desc_struct = struct.Struct('>Q')
    crc32_struct = struct.Struct('>I')
    leaf_struct = struct.Struct('>QQ')

    def __init__(self, store, header, order=None, key_type=None,
                 value_type=None, compress=None):
        """Create provider

        Creates new provider or loads existing one identified by name. Compress
        argument specifies compression level default is 9(maximum compression).
        If key_type(value_type) is not specified they are set to 'pickle'
        """
        self.store = store
        self.header = header
        self.d2n = {}
        self.desc_next = -1
        self.dirty_set = set()

        header_data = header()
        if header_data:
            state_json = header_data[:-self.crc32_struct.size]
            crc32 = self.crc32_struct.unpack(header_data[-self.crc32_struct.size:])[0]
            if crc32 != binascii.crc32(state_json) & 0xffffffff:
                raise ValueError('Header checksum failed')
            state = json.loads(state_json.decode())
            self.size_st = state['size']
            self.depth_st = state['depth']
            self.order_st = state['order']
            self.key_type, self.read_keys, self.write_keys = self.type_parse(state['key_type'])
            self.value_type, self.read_vals, self.write_vals = self.type_parse(state['value_type'])
            self.compress = state.get('compress', 0)
            self.root_st = self.node_load(state['root'])

        else:
            self.size_st = 0
            self.depth_st = 1
            self.order_st = order or self.order_default
            self.key_type, self.read_keys, self.write_keys = self.type_parse(key_type or self.type_default)
            self.value_type, self.read_vals, self.write_vals = self.type_parse(value_type or self.type_default)
            self.compress = compress if compress is not None else self.compress_default
            self.root_st = self.create([], [], True)

    def flush(self, prune=None):
        """Flush provider and store
        """
        d2n_reloc = {}  # relocated nodes
        leaf_queue = {}

        ## flush leafs
        def leaf_enqueue(leaf):
            leaf_stream = io.BytesIO()

            # save leaf
            leaf_stream.seek(self.leaf_struct.size)
            if self.compress:
                with CompressorStream(leaf_stream, self.compress) as stream:
                    self.read_keys(stream, leaf.keys)
                    self.read_vals(stream, leaf.children)
            else:
                self.read_keys(leaf_stream, leaf.keys)
                self.read_vals(leaf_stream, leaf.children)

            # enqueue leaf
            leaf_queue[leaf] = leaf_stream

            # allocate space
            desc = self.store.reserve(leaf_stream.tell() + 1, None if leaf.desc < 0 else leaf.desc)
            if leaf.desc != desc:
                # queue parent for update
                if leaf is not self.root_st:
                    parent, key = self.root_st, leaf.keys[0]
                    while True:
                        parent_desc = parent.children[bisect(parent.keys, key)]
                        if parent_desc == leaf.desc:
                            break
                        parent = self.desc_to_node(parent_desc)
                    if parent not in self.dirty_set:
                        node_queue.add(parent)

                # queue next and previous for update
                for sibling_desc in(leaf.prev, leaf.next):
                    # descriptor is negative, node is dirty
                    if(sibling_desc > 0 and                # negative node is dirty for sure
                       sibling_desc not in d2n_reloc):     # relocated node is also dirty

                        sibling = self.d2n.get(sibling_desc)
                        if sibling:
                            # node has already been loaded
                            if(sibling not in self.dirty_set and
                               sibling not in leaf_queue):
                                    # queue it for update
                                    leaf_enqueue(sibling)
                        else:
                            # node hasn't been loaded
                            leaf_enqueue(self.node_load(sibling_desc))

                # update descriptor maps
                self.d2n.pop(leaf.desc)
                d2n_reloc[leaf.desc], leaf.desc = leaf, desc
                self.d2n[desc] = leaf

        # enqueue leafs and create dirty nodes queue
        node_queue = set()
        for node in self.dirty_set:
            if node.is_leaf:
                leaf_enqueue(node)
            else:
                node_queue.add(node)

        # all leafs has been allocated now
        for leaf, leaf_stream in leaf_queue.items():
            # update previous
            prev = d2n_reloc.get(leaf.prev)
            if prev is not None:
                leaf.prev = prev.desc
            # update next
            next = d2n_reloc.get(leaf.next)
            if next is not None:
                leaf.next = next.desc

            # leaf header(perv, next)
            leaf_stream.seek(0)
            leaf_stream.write(self.leaf_struct.pack(leaf.prev, leaf.next))

            # leaf tag
            leaf_stream.seek(0, io.SEEK_END)
            leaf_stream.write(b'\x01')

            # put leaf in store
            desc = self.store.save(leaf_stream.getvalue(), leaf.desc)
            assert leaf.desc == desc

        ## flush nodes
        def node_flush(node):
            # flush children
            for index in range(len(node.children)):
                child_desc = node.children[index]
                child = d2n_reloc.get(child_desc)
                if child is not None:
                    # child has already been flushed
                    node.children[index] = child.desc
                else:
                    child = self.d2n.get(child_desc)
                    if child in node_queue:
                        # flush child and update index
                        node.children[index] = node_flush(child)

            # node
            node_stream = io.BytesIO()
            if self.compress:
                with CompressorStream(node_stream, self.compress) as stream:
                    self.read_keys(stream, node.keys)
                    Serializer(stream).write_struct_list(node.children, self.desc_struct)
            else:
                self.read_keys(node_stream, node.keys)
                Serializer(node_stream).write_struct_list(node.children, self.desc_struct)

            # node tag
            node_stream.write(b'\x00')

            # put node in store
            desc = self.store.save(node_stream.getvalue(), None if node.desc < 0 else node.desc)

            # check if node has been relocated
            if node.desc != desc:
                # queue parent for update
                if node is not self.root_st:
                    parent, key = self.root_st, node.keys[0]
                    while True:
                        parent_desc = parent.children[bisect(parent.keys, key)]
                        if parent_desc == node.desc:
                            break
                        parent = self.d2n[parent_desc]
                    if parent not in self.dirty_set:
                        node_queue.add(parent)

                # update descriptor maps
                self.d2n.pop(node.desc)
                d2n_reloc[node.desc], node.desc = node, desc
                self.d2n[desc] = node

            # remove node from dirty set
            node_queue.discard(node)

            return desc

        while node_queue:
            node_flush(node_queue.pop())

        # clear dirty
        self.dirty_set.clear()
        if prune:
            self.d2n.clear()  # release all nodes except root_st
            self.d2n[self.root_st.desc] = self.root_st

        state = {
            'size': self.size_st,
            'depth': self.depth_st,
            'order': self.order_st,
            'key_type': self.key_type,
            'value_type': self.value_type,
            'compress': self.compress,
            'root': self.root_st.desc
        }
        state_json = json.dumps(state, sort_keys=True).encode()
        crc32 = binascii.crc32(state_json) & 0xffffffff

        header_data = state_json + self.crc32_struct.pack(crc32)
        if self.header() != header_data:
            self.header(header_data)

    def size(self, value=None):
        self.size_st = self.size_st if value is None else value
        return self.size_st

    def depth(self, value=None):
        self.depth_st = self.depth_st if value is None else value
        return self.depth_st

    def order(self):
        return self.order_st

    def root(self, value=None):
        self.root_st = self.root_st if value is None else value
        return self.root_st

    def size_in_store(self):
        """Size occupied on store
        """
        return (functools.reduce(operator.add, (StoreBlock.from_desc(node.desc).size
                for node in self if node.desc > 0), 0))

    def node_to_desc(self, node):
        return node.desc

    def desc_to_node(self, desc):
        if desc:
            return self.d2n.get(desc) or self.node_load(desc)

    def create(self, keys, children, is_leaf):
        desc, self.desc_next = self.desc_next, self.desc_next - 1
        node = (StoreBPTreeLeaf(desc, keys, children) if is_leaf else
                StoreBPTreeNode(desc, keys, children))
        self.d2n[desc] = node
        self.dirty_set.add(node)
        return node

    def dirty(self, node):
        self.dirty_set.add(node)

    def release(self, node):
        self.d2n.pop(node.desc)
        self.dirty_set.discard(node)
        if node.desc >= 0:
            self.store.delete(node.desc)

    def drop(self):
        """Completely delete provider from the store
        """
        self.flush()
        for node in tuple(self):
            self.store.delete(node.desc)
        self.header(b'')
        self.size_st = 0
        self.depth_st = 1
        self.root_st = self.create([], [], True)

    def node_load(self, desc):
        """Load node by its descriptor
        """
        node_data = self.store.load(desc)
        node_tag = node_data[-1:]

        if node_tag != b'\x01':
            node_stream = (io.BytesIO(node_data[:-1]) if not self.compress else
                           io.BytesIO(zlib.decompress(node_data[:-1])))
            node = StoreBPTreeNode(desc,
                                   self.write_keys(node_stream),
                                   Serializer(node_stream).read_struct_list(self.desc_struct))
        else:
            prev, next = self.leaf_struct.unpack(node_data[:self.leaf_struct.size])
            node_stream = (io.BytesIO(node_data[self.leaf_struct.size:-1]) if not self.compress else
                           io.BytesIO(zlib.decompress(node_data[self.leaf_struct.size:-1])))
            node = StoreBPTreeLeaf(desc,
                                   self.write_keys(node_stream),
                                   self.write_vals(node_stream))
            node.prev = prev
            node.next = next

        self.d2n[desc] = node
        return node

    def type_parse(self, type):
        """Parse type

        Returns(type, to_stream, from_steram) for specified type.
        """
        if type == 'bytes':
            return('bytes',
                   lambda stream, items: Serializer(stream).write_bytes_list(items),
                   lambda stream: Serializer(stream).read_bytes_list())

        elif type.startswith('pickle'):
            protocol = int(type.partition(':')[-1] or str(pickle.HIGHEST_PROTOCOL))
            return('pickle:{}'.format(protocol),
                   lambda stream, items: pickle.dump(items, stream, protocol),
                   lambda stream: pickle.load(stream))

        elif type.startswith('struct:'):
            format = type.partition(':')[-1].encode()
            it_struct = struct.Struct(format)
            it_comp = len(format.translate(None, b'<>=!@')) > 1
            return('struct:{}'.format(format.decode()),
                   lambda stream, items: Serializer(stream).write_struct_list(items, it_struct, it_comp),
                   lambda stream: Serializer(stream).read_struct_list(it_struct, it_comp))

        elif type == 'json':
            encode = codecs.getencoder('utf-8')
            decode = codecs.getdecoder('utf-8')
            header = struct.Struct('>Q')
            header_size = header.size

            def json_save(stream, items):
                data = encode(json.dumps(items))[0]
                stream.write(header.pack(len(data)))
                stream.write(data)

            def json_load(stream):
                data_size = header.unpack(stream.read(header_size))[0]
                return json.loads(decode(stream.read(data_size))[0])

            return('json', json_save, json_load)

        raise ValueError('Unknown serializer type: {}'.format(type))


class StoreBPTreeNode(BPTreeNode):
    """Store B+Tree Node
    """
    __slots__ = BPTreeNode.__slots__ + ('desc',)

    def __init__(self, desc, keys, children):
        self.desc = desc
        self.keys = keys
        self.children = children
        self.is_leaf = False


class StoreBPTreeLeaf(BPTreeLeaf):
    """Store B+Tree Leaf
    """
    __slots__ = BPTreeLeaf.__slots__ + ('desc',)

    def __init__(self, desc, keys, children):
        self.desc = desc
        self.keys = keys
        self.children = children
        self.prev = 0
        self.next = 0
        self.is_leaf = True


class CompressorStream(object):
    """Compression stream adapter
    """
    __slots__ = ('stream', 'compressor',)

    def __init__(self, stream, level):
        self.stream = stream
        self.compressor = zlib.compressobj(level)

    def write(self, data):
        return self.stream.write(self.compressor.compress(data))

    def __enter__(self):
        return self

    def __exit__(self, et, oe, tb):
        self.stream.write(self.compressor.flush())
        return False



"""Message hub
"""
import threading
import itertools
from ..monad import Cont, async_block

__all__ = ('Hub', 'Sender', 'Receiver', 'pair',)


class Hub(object):
    """Message hub
    """
    inst_lock = threading.RLock()
    inst_local = threading.local()
    inst_main = None

    def __init__(self):
        self.handlers = {}
        addr_iter = itertools.count(1)
        self.addr = lambda: Address((next(addr_iter),))

    @classmethod
    def main(cls, inst=None):
        """Main core instance
        """
        with cls.inst_lock:
            if inst is None:
                if cls.inst_main is None:
                    cls.inst_main = Hub()
                inst = cls.inst_main
            else:
                cls.inst_main = inst
        return inst

    @classmethod
    def local(cls, inst=None):
        """Thread local instance
        """
        if inst is None:
            inst = getattr(cls.inst_local, 'inst', None)
            if inst is None:
                return cls.local(cls.main())
        else:
            cls.inst_local.inst = inst
        return inst

    def send(self, msg, dst, src):
        if not self.try_send(msg, dst, src):
            raise ValueError('no receiver for address: {}'.format(dst))

    def try_send(self, msg, dst, src):
        handler = self.handlers.get(dst, None)
        if handler is None:
            return False
        else:
            if not handler(msg, dst, src):
                self.handlers.pop(dst)
            return True

    def recv(self, dst, handler):
        if self.handlers.setdefault(dst, handler) != handler:
            raise ValueError('multiple receive handlers for address: {}'.format(dst))
        return handler

    def recv_once(self, dst, handler):
        def once_handler(msg, dst, src):
            handler(msg, dst, src)
            return False
        return self.recv(dst, once_handler)

    def unrecv(self, dst):
        return self.handlers.pop(dst, None) is not None

    def __len__(self):
        return len(self.handlers)

    def __str__(self):
        return 'Hub(len:{})'.format(len(self))

    def __repr__(self):
        return str(self)


class Address(tuple):
    """Hub address
    """
    __slots__ = tuple()

    def route(self, addr):
        return Address(self + addr)

    def unroute(self):
        if not self:
            raise ValueError('empty address cannot be unrouted')
        return Address(self[:-1])

    def __eq__(self, addr):
        return self[-1] == addr[-1]

    def __hash__(self):
        return hash(self[-1])

    def __str__(self):
        return '.'.join(str(p) for p in reversed(self))

    def __repr__(self):
        return str(self)


class Sender(object):
    """Hub sender
    """
    __slots__ = ('hub', 'addr',)

    def __init__(self, hub, addr):
        self.hub = hub
        self.addr = addr

    def send(self, msg, src=None):
        return self.hub.send(msg, self.addr, src)

    def try_send(self, msg, src=None):
        return self.hub.try_send(msg, self.addr, src)

    def __call__(self, msg):
        @async_block
        def sender_cont(ret):
            addr = self.hub.addr()
            self.hub.recv_once(addr, lambda msg, dst, src: ret(msg))
            try:
                self.hub.send(msg, self.addr, Sender(self.hub, addr))
            except Exception:
                self.hub.unrecv(addr)
                raise
        return sender_cont

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                (self.hub, self.addr) == (other.hub, other.addr))

    def __hash__(self):
        return hash(self.addr)

    def __str__(self):
        return 'Sender(addr:{})'.format(self.addr)

    def __repr__(self):
        return str(self)


class Receiver(object):
    """Hub receiver
    """
    __slots__ = ('hub', 'addr',)

    def __init__(self, hub, addr):
        if len(addr) > 1:
            raise ValueError('non local address {}'.format(addr))
        self.hub = hub
        self.addr = addr

    def future(self):
        return self.__monad__().future()

    def __call__(self, handler):
        return self.hub.recv(self.addr, handler)

    def __monad__(self):
        return Cont(lambda ret: self.hub.recv_once(self.addr,
                    lambda msg, dst, src: ret(msg, src)))

    def dispose(self):
        self.hub.unrecv(self.addr)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return ('Receiver(addr:{}, handler:{})'.format(self.addr,
                getattr(self.hub.handlers.get(self.addr), '__name__', None)))

    def __repr__(self):
        return str(self)


def pair(addr=None, hub=None):
    """Create receiver-sender pair
    """
    hub = hub or Hub.local()
    addr = addr or hub.addr()
    return Receiver(hub, addr), Sender(hub, addr)



"""Identity monad implementation
"""
from .monad import Monad

__all__ = ('Identity',)


class Identity(Monad):
    """Identity monad
    """
    __slots__ = ('value',)

    def __init__(self, value):
        self.value = value

    def bind(self, func):
        return func(self.value)

    @classmethod
    def unit(cls, val):
        return cls(val)

    def __str__(self):
        return 'Identity({})'.format(self.value)

    def __repr__(self):
        return str(self)



import functools
from .. import PRETZEL_TEST_TIMEOUT
from ..core import Core, sleep
from ..uniform import CanceledError
from ..monad import async
from ..remoting.hub import Hub

__all__ = ('async_test',)


def async_test(test):
    """Run asynchronous test in context of newly create core object
    """
    @functools.wraps(test)
    def test_async(*args, **kwargs):
        core_prev = Core.local()
        try:
            with Core.local(Core()) as core:
                @async
                def timeout():
                    yield core.sleep(PRETZEL_TEST_TIMEOUT)
                    raise CanceledError('test timeout')

                test_future = (async(test)(*args, **kwargs) | timeout()).future()
                test_future(lambda _: core.dispose())
                if not core.disposed:
                    core()
        finally:
            Core.local(core_prev)
            Hub.local(Hub())
        assert test_future.completed
        test_future.value
    return test_async


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import event, dispose, process, task, boot, utils

    suite = TestSuite()
    for test in (event, dispose, process, task, boot, utils):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



"""Stream connection
"""
from .conn import Connection
from ...monad import async
from ...uniform import CanceledError, BrokenPipeError

__all__ = ('StreamConnection',)


class StreamConnection(Connection):
    """Stream based connected
    """
    def __init__(self, hub=None, core=None):
        Connection.__init__(self, hub=hub, core=core)
        self.reader = None
        self.writer = None

    @async
    def do_connect(self, target):
        """Connect implementation

        Target is tuple of input and output streams.
        """
        @async
        def recv_coro():
            try:
                if self.reader.disposed:
                    return
                # Begin read next message before dispatching current one, as
                # connection may be closed during dispatching and input stream
                # became disposed.
                msg_next = self.reader.read_bytes().future()
                while True:
                    msg, msg_next = (yield msg_next), self.reader.read_bytes().future()
                    self.do_recv(msg)()
            except (CanceledError, BrokenPipeError):
                pass
            finally:
                self.dispose()

        self.reader, self.writer = target
        recv_coro()()

    def do_disconnect(self):
        if self.reader is not None:
            self.reader.dispose()
        if self.writer is not None:
            self.writer.dispose()

    def do_send(self, msg):
        self.writer.write_bytes(msg)
        self.writer.flush()()
        return True



import unittest

from ..expr import *
from ...event import Event
from ...monad import Identity

__all__ = ('ExprTest',)


class ExprTest(unittest.TestCase):
    def test_load_arg(self):
        self.assertEqual(run_expr(LoadArgExpr(0), 'one', 'two'), 'one')
        self.assertEqual(run_expr(LoadArgExpr(1), 'one', 'two'), 'two')

    def test_load_const(self):
        self.assertEqual(run_expr(LoadConstExpr('constant')), 'constant')

    def test_call(self):
        fn = lambda *a, **kw: (a, kw)
        self.assertEqual(run_expr(CallExpr(fn, 'arg')), fn('arg'))
        self.assertEqual(run_expr(CallExpr(fn, 0, 1, 2, one=1, two=2)),
                         fn(0, 1, 2, one=1, two=2))
        self.assertEqual(run_expr(CallExpr(fn, one=1, two=2)), fn(one=1, two=2))

    def test_attr(self):
        class A(object):
            pass
        a = A()

        get_attr = GetAttrExpr(LoadConstExpr(a), 'key').code()
        set_attr = SetAttrExpr(LoadConstExpr(a), 'key', LoadArgExpr(0)).code()

        with self.assertRaises(AttributeError):
            run(get_attr)
        self.assertEqual(run(set_attr, 'value'), None)
        self.assertEqual(a.key, 'value')
        self.assertEqual(run(get_attr), 'value')

    def test_item(self):
        d = {}
        get_item = GetItemExpr(LoadConstExpr(d), LoadArgExpr(0)).code()
        set_item = SetItemExpr(LoadConstExpr(d), LoadArgExpr(0), LoadArgExpr(1)).code()

        with self.assertRaises(KeyError):
            run(get_item, 'key')
        self.assertEqual(run(set_item, 'key', 'value'), None)
        self.assertEqual({'key': 'value'}, d)
        self.assertEqual(run(get_item, 'key'), 'value')

    def test_raise(self):
        with self.assertRaises(RuntimeError):
            run_expr(RaiseExpr(LoadArgExpr(0)), RuntimeError('test'))

    def test_bind(self):
        ev = Event()
        ev_future = CallExpr(len, BindExpr(LoadArgExpr(0))).code()(ev).future()

        self.assertFalse(ev_future.completed)
        ev('result')
        self.assertEqual(ev_future.value, len('result'))

    def test_if(self):
        max_code = IfExpr(CmpExpr('>', LoadArgExpr(0), LoadArgExpr(1)),
                          LoadArgExpr(0),
                          LoadArgExpr(1)).code()

        self.assertEqual(run(max_code, 3, 2), 3)
        self.assertEqual(run(max_code, 2, 3), 3)

    def test_while(self):
        while_code = WhileExpr(CmpExpr('>', CallExpr(LoadArgExpr(0)), LoadConstExpr(0)),
                               CallExpr(LoadArgExpr(1))).code()
        ctx = [10, 0]

        def cond():
            ctx[0] -= 2
            return ctx[0]

        def body():
            ctx[1] += 1

        run(while_code, cond, body)
        self.assertEqual(ctx, [0, 4])


def run(code, *args):
    return code(*args, monad=Identity).value.value


def run_expr(expr, *args):
    return run(expr.code(), *args)



import time
import unittest
from . import async_test
from ..monad import async_all
from ..task import ThreadPool, task

__all__ = ('TaskTest',)


class TaskTest(unittest.TestCase):
    @async_test
    def test(self):
        """This test may fail if system is really busy
        """
        sleep_time = .3
        sleep_action = lambda: time.sleep(sleep_time)
        try:
            with ThreadPool.main() as pool:
                # thread count test
                start = time.time()
                yield async_all((task(sleep_action),) * (pool.size + 1))
                stop = time.time()
                self.assertEqual(int((stop - start) / sleep_time), 2)

                # error test
                def error():
                    raise RuntimeError()
                with self.assertRaises(RuntimeError):
                    yield task(error)

                # disposed task
                pool.size = 1
                task(sleep_action)(lambda _: None)
                disp_task = task(lambda: time.sleep(100)).future()

            with self.assertRaises(ValueError):
                yield disp_task

        finally:
            ThreadPool.main(None)



"""In memory B+Tree provider

Mainly for testing purposes.
"""
from .provider import BPTreeProvider
from ..bptree import BPTreeNode, BPTreeLeaf

__all__ = ('MemoryBPTreeProvider',)


class MemoryBPTreeProvider(BPTreeProvider):
    """In memory B+Tree Provider
    """
    def __init__(self, order):
        self.root_st = self.create([], [], True)
        self.size_st = 0
        self.depth_st = 1
        self.order_st = order

    def size(self, value=None):
        if value is not None:
            self.size_st = value
        return self.size_st

    def depth(self, value=None):
        if value is not None:
            self.depth_st = value
        return self.depth_st

    def order(self):
        return self.order_st

    def root(self, value=None):
        if value is not None:
            self.root_st = value
        return self.root_st

    def node_to_desc(self, node):
        return node

    def desc_to_node(self, desc):
        return desc

    def create(self, keys, children, is_leaf):
        return (BPTreeLeaf(keys, children) if is_leaf else
                BPTreeNode(keys, children))

    def dirty(self, node):
        pass

    def release(self, node):
        pass



import pickle
import unittest
from ..closure import Closure

__all__ = ('ClosureTest',)


class ClsoureTest(unittest.TestCase):
    def test(self):
        reload = lambda val: pickle.loads(pickle.dumps(val))

        # reference noting
        self.assertEqual(reload(Closure(lambda val: val))('id'), 'id')

        # reference global
        self.assertEqual(reload(Closure(lambda: GLOBAL))(), GLOBAL)

        # reference closure
        cl = reload(Closure((lambda a: lambda b: a + b)('closure')))
        self.assertEqual(cl(':arg'), 'closure:arg')

GLOBAL = 'test value'



"""Asynchronous application trampoline decorator
"""
from functools import wraps
from .core import Core
from .monad import async, async_green

__all__ = ('app', 'app_green', 'app_run',)


def app(main):
    """Application decorator

    Create application from generator function.
    """
    return wraps(main)(lambda *a, **kw: app_run(async(main)(*a, **kw)))


def app_green(main):
    """Application decorator

    Create application from greenlet main function.
    """
    return wraps(main)(lambda *a, **kw: app_run(async_green(main)(*a, **kw)))


def app_run(cont):
    """Application trampoline

    Run continuation as application.
    """
    with Core.local(Core()) as core:
        app_future = cont.__monad__().future()
        app_future(lambda _: core.dispose())
        if not core.disposed:
            core()
    assert app_future.completed
    return app_future.value



import io
import socket
import unittest
import itertools
from ..sock import Socket
from ...monad import async
from ...uniform import BrokenPipeError
from ...core import schedule
from ...tests import async_test

__all__ = ('SockTest',)


class SockTest(unittest.TestCase):
    PORT = 54321
    BLOCK = b'X' * 1024

    @async_test
    def test(self):
        received = io.BytesIO()

        @async
        def server_coro():
            with Socket(socket.socket()) as sock:
                sock.bind(('localhost', self.PORT))
                sock.listen(10)
                client, addr = yield sock.accept()
                with client:
                    try:
                        while True:
                            received.write((yield client.read(1024)))
                    except BrokenPipeError:
                        pass
        server = server_coro().future()
        server.__monad__()()  # error trace if any
        if server.completed:
            yield server

        with Socket(socket.socket()) as sock:
            yield sock.connect(('localhost', self.PORT))

            # fill socket buffer
            size = 0
            for _ in itertools.count():
                write = sock.write(self.BLOCK).future()
                if not write.completed:
                    size += yield write
                    break
                else:
                    size += write.value

            # drain socket
            while received.tell() < size:
                yield schedule()
            self.assertEqual(received.getvalue(), b'X' * size)
            received.seek(0)
            received.truncate(0)

            # send some data
            yield sock.write(b'one')
            yield schedule()

            yield sock.write(b', two')
            yield schedule()

            yield sock.write(b', three')
            yield schedule()

        yield server
        self.assertEqual(received.getvalue(), b'one, two, three')



from . import core, poll
from .core import *
from .poll import *

__all__ = (core.__all__ + poll.__all__ + ('sleep', 'sleep_until',
           'poll', 'schedule',))


def sleep(delay, core=None):
    """Sleep for "delay" seconds
    """
    return (core or Core.local()).sleep(delay)


def sleep_until(when, core=None):
    """Sleep until "when" time is reached
    """
    return (core or Core.local()).sleep_until(when)


def poll(fd, mask, core=None):
    """Poll file descriptor

    Poll file descriptor for events specified by mask. If mask is None then
    specified descriptor is unregistered and all pending events are resolved
    with BrokenPipeError, otherwise future is resolved with bitmap of the
    events happened on file descriptor or error if any.
    """
    return (core or Core.local()).poll(fd, mask)


def schedule(core=None):
    """Schedule continuation to be executed on specified core (local by default)

    Scheduled continuation will be executed on next iteration circle. This
    function can be called from different thread.
    """
    return (core or Core.local()).schedule()


def waitpid(pid, core=None):
    """Wait pid

    Schedule continuation to be executed when process with pid is terminated.
    """
    return (core or Core.local()).waitpid(pid)


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import tests

    suite = TestSuite()
    for test in (tests,):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



# -*- coding: utf-8 -*-
import io
import struct
import unittest
from ..serialize import Serializer

__all__ = ('SerializerTest',)


class SerializerTest(unittest.TestCase):
    """Serializer unit test
    """
    def test_struct(self):
        """Structure serializer
        """
        struct_save = []
        format = struct.Struct('>L')

        # empty
        stream = io.BytesIO()
        serial = Serializer(stream)
        serial.write_struct_list(struct_save, format)

        stream.seek(0)
        struct_load = serial.read_struct_list(format)
        self.assertEqual(struct_save, struct_load)

        # normal
        stream = io.BytesIO()
        serial = Serializer(stream)
        struct_save.extend(range(10))
        serial.write_struct_list(struct_save, format)

        stream.seek(0)
        struct_load = serial.read_struct_list(format)
        self.assertEqual(struct_save, struct_load)

        # tuple
        stream = io.BytesIO()
        serial = Serializer(stream)
        struct_save = [(i, i) for i in range(10)]
        format = struct.Struct('BB')
        serial.write_struct_list(struct_save, format, True)

        stream.seek(0)
        struct_load = serial.read_struct_list(format, True)
        self.assertEqual(struct_save, struct_load)

    def test_bytes(self):
        """Bytes serializer
        """
        bytes_save = []

        # empty
        stream = io.BytesIO()
        serial = Serializer(stream)
        serial.write_bytes_list(bytes_save)

        stream.seek(0)
        bytes_load = serial.read_bytes_list()
        self.assertEqual(bytes_save, bytes_load)

        # normal
        stream = io.BytesIO()
        serial = Serializer(stream)
        bytes_save.extend(str(i).encode() for i in range(10))
        serial.write_bytes_list(bytes_save)

        stream.seek(0)
        bytes_load = serial.read_bytes_list()
        self.assertEqual(bytes_save, bytes_load)



from . import store
from .store import *

__all__ = store.__all__


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import tests

    suite = TestSuite()
    for test in (tests,):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



import io
import unittest
from ..store import StreamStore

__all__ = ('StoreStreamTest',)


class StoreStreamTest(unittest.TestCase):
    def testCheck(self):
        store = StreamStore(io.BytesIO())
        stream = store.create_stream(b'test', bufsize=8)

        # read outside of data
        self.assertEqual(stream.seek(10), 10)
        self.assertEqual(stream.read(), b'')
        self.assertEqual(stream.size, 0)

        # write outside of data(zero filling)
        self.assertEqual(stream.write(b'X'), 1)
        self.assertEqual(stream.seek(0), 0)
        self.assertEqual(stream.size, 11)
        self.assertEqual(stream.read(), b'\x00' * 10 + b'X')
        self.assertEqual(stream.tell(), 11)

        # data update
        stream.seek(1)
        self.assertEqual(stream.write(b'ABC'), 3)
        self.assertEqual(stream.read(1), b'\x00')
        stream.seek(0)
        self.assertEqual(stream.read(5), b'\x00ABC\x00')
        stream.flush()

        # reload stream
        stream = store.create_stream(b'test')
        self.assertEqual(stream.read(), b'\x00ABC' + b'\x00' * 6 + b'X')

    def test_mult(self):
        store = StreamStore(io.BytesIO())
        with store.create_stream('a') as a, store.create_stream('b') as b:
            a.write(b'stream a')
            b.write(b'stream b')
        with store.create_stream('a') as a, store.create_stream('b') as b:
            self.assertEqual(a.read(), b'stream a')
            self.assertEqual(b.read(), b'stream b')



import unittest
from ..hub import Hub
from ..proxy import Proxy, proxify
from ...event import Event
from ...monad import async, do_return
from ...core import schedule
from ...tests import async_test

__all__ = ('ProxyTest',)


class ProxyTest(unittest.TestCase):
    @async_test
    def test(self):
        remote = Remote('val')
        with proxify(remote) as proxy:
            # simple
            self.assertEqual((yield proxy), remote)
            self.assertEqual((yield proxy('test')), 'test')

            # attribute
            self.assertEqual((yield proxy.value), 'val')
            with self.assertRaises(AttributeError):
                yield proxy.bad_attr

            # item
            self.assertEqual((yield proxy['item']), 'item_value')
            with self.assertRaises(KeyError):
                yield proxy['bad item']

            # method
            self.assertEqual((yield proxy.method()), (yield proxy.value))
            self.assertEqual((yield proxy.method('val_new')), 'val')
            self.assertEqual((yield proxy.method()), (yield proxy.value))
            with self.assertRaises(RuntimeError):
                yield proxy.method_error(RuntimeError())

            # asynchronous method
            async_val = (~proxy.method_async()).__monad__().future()
            self.assertFalse(async_val.completed)
            yield proxy()
            self.assertEqual((yield async_val).value, remote.value)

            with (yield proxify(proxy.value)) as value_proxy:
                self.assertTrue(isinstance(value_proxy, Proxy))

        yield schedule()  # handlers will be cleaned when coroutine is interrupted
        self.assertFalse(len(Hub.local()), 0)


class Remote (object):
    def __init__(self, value):
        self.items = {'item': 'item_value'}
        self.value = value
        self.event = Event()

    def __getitem__(self, name):
        return self.items[name]

    def __setitem__(self, name, value):
        self.items[name] = value

    def method(self, value=None):
        if value is None:
            return self.value
        else:
            value, self.value = self.value, value
            return value

    @async
    def method_async(self, value=None):
        yield self.event
        do_return(self.method(value))

    def method_error(self, error):
        raise error

    def __call__(self, value=None):
        self.event(value)
        return value

    def __monad__(self):
        self.event.__monad__().map_val(lambda r: r[0])



from .bptree import BPTree
from .provider.store import StoreBPTreeProvider

__all__ = ('StoreMap',)


class StoreMap(BPTree):
    """B+Tree with store as back end
    """
    def __init__(self, store, header, order=None, key_type=None, value_type=None, compress=None):
        BPTree.__init__(self, StoreBPTreeProvider(store, header, order, key_type, value_type, compress))

    @property
    def header(self):
        return self.provider.header

    @property
    def store(self):
        return self.provider.store

    @property
    def size(self):
        """Size occupied on store
        """
        return self.provider.size_in_store()

    def flush(self, prune=None):
        """Flush dirty nodes to store
        """
        self.provider.flush(prune)

    def drop(self):
        """Completely delete mapping from the store
        """
        self.provider.drop()

    def dispose(self):
        """Flush dirty nodes to store
        """
        self.flush(prune=True)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return ('StoreMap(len:{}, size:{}, key_type:{}, val_type:{}, compress:{})'
                .format(len(self), self.size, self.provider.key_type,
                        self.provider.value_type, self.provider.compress))

    def __repr__(self):
        return str(self)



"""Different kinds of monads and do block notation
"""
from . import monad, ident, result, list, cont, proxy
from . import async as _async
from . import do as _do
from . import do_green as _do_green
from .monad import *
from .do import *
from .do_green import *
from .ident import *
from .result import *
from .list import *
from .cont import *
from .async import *
from .proxy import *

__all__ = (monad.__all__ + _do.__all__ + _do_green.__all__ + ident.__all__ +
           result.__all__ + list.__all__ + cont.__all__ + _async.__all__ +
           proxy.__all__)


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import tests

    suite = TestSuite()
    for test in (tests,):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



import os
from distutils.core import setup


def package_full(package):
    packages = []
    for path, dirs, files in os.walk(package):
        for file in files:
            if file != '__init__.py':
                continue
            packages.append('.'.join(name for name in path.split(os.path.sep)
                                     if name))
    return packages

setup(name='pretzel',
      version='1.0.5',
      author='Pavel Aslanov',
      description='Pretzel asynchronous python framework',
      long_description=open('README.rst').read(),
      url='https://github.com/aslpavel/pretzel',
      packages=package_full('pretzel'))



import io
import sys
import unittest
from ..boot import BootLoader

__all__ = ('BootTest',)


class BootTest(unittest.TestCase):
    def test_loader(self):
        # simple module
        name = '_boot_test'
        loader = BootLoader(name, 'VALUE="value"', '<stdin>', False)

        self.assertFalse(loader.is_package(name))
        self.assertEqual(loader.get_source(name), 'VALUE="value"')
        self.assertTrue(isinstance(loader.get_code(name),
                                   type((lambda: None).__code__)))

        name_bad = name + '_bad'
        for action in (lambda: loader.load_module(name_bad),
                       lambda: loader.is_package(name_bad),
                       lambda: loader.get_source(name_bad),
                       lambda: loader.get_code(name_bad)):
            with self.assertRaises(ImportError):
                action()

        mod = loader()
        self.assertTrue(mod is loader())
        try:
            self.assertTrue(name in sys.modules)
            self.assertEqual(mod.VALUE, 'value')
            self.assertEqual(mod.__file__, '<stdin>')
            self.assertEqual(mod.__name__, name)
            self.assertEqual(mod.__package__, '')
            self.assertFalse(hasattr(mod, '__path__'))
        finally:
            sys.modules.pop(mod.__name__)

        # restore
        loader_stream = loader.to_stream(io.BytesIO())
        loader_stream.seek(0)
        loader_restored = BootLoader.from_stream(loader_stream)
        self.assertFalse(loader_restored.is_package(name))
        self.assertEqual(loader_restored.get_source(name),
                         loader.get_source(name))

        # bad module
        bad_loader = BootLoader('_boot_bad_test', 'raise RuntimeError()',
                                '<stdin>', False)
        with self.assertRaises(RuntimeError):
            bad_loader()
        self.assertTrue('_boot_bad_test' not in sys.modules)



"""Asynchronous pretzel shell

Requires greenlet module to be installed.
"""
import os
import sys
import code
import signal
import socket
import textwrap
import functools
from collections import deque
from ..utils import lazy
from ..monad import Result, async, async_any, async_green, bind_green
from ..core import Core, schedule
from ..event import Event
from ..stream import BufferedFile
from ..uniform import BrokenPipeError, CanceledError
from ..remoting import ForkConnection, SSHConnection, pair
from ..state_machine import StateMachine
from .. import __package__ as pretzel_pkg

__all__ = ('Shell', 'ShellStream', 'prompt',)


class ShellBase(object):
    OP_EXEC = 0
    OP_PUSH = 1
    OP_INT = 2
    OP_EOT = 3

    def __init__(self, sender):
        self.sender = sender

    def __call__(self, locals=None, banner=None):
        """Execute shell
        """
        return self.sender((self.OP_EXEC, (locals, banner)))

    def push(self, line):
        """Push line to shell
        """
        return self.sender.send((self.OP_PUSH, line))

    def interrupt(self):
        """Interrupt shell
        """
        return self.sender.send((self.OP_INT, None))

    def end_of_transmission(self):
        """End-of-transmission
        """
        return self.sender.try_send((self.OP_EOT, None))

    def __reduce__(self):
        return ShellBase, (self.sender,)

    def __repr__(self):
        return 'Shell(addr:{})'.format(self.sender.addr)


class Shell(ShellBase):
    BANNER = textwrap.dedent("""\
        Welcome to asynchronous pretzel shell!
            {}
            bind   - bind value associated with continuation monad
            bound  - create bound version of asynchronous function
            monad  - get associated monad
            attach - attach shell to specified connection
            fork   - create fork connection
            ssh    - create ssh connection
            core   - application's core object

        -> host:{{host}} pid:{{pid}}""").format('')

    INIT_SOURCE = textwrap.dedent("""\
        from __future__ import print_function
        from {pkg}.core import *
        from {pkg}.monad import *
        from {pkg}.event import *
        from {pkg}.stream import *
        from {pkg}.process import *
        from {pkg}.task import *
        from {pkg}.remoting import *
        """.format(pkg=pretzel_pkg))

    STATE_INIT = 0
    STATE_EXEC = 1
    STATE_DISP = 2
    STATE_GRAPH = StateMachine.compile_graph({
        STATE_INIT: (STATE_EXEC, STATE_DISP),
        STATE_EXEC: (STATE_DISP,),
        STATE_DISP: (STATE_DISP,),
    })
    STATE_NAMES = ('initial', 'executing', 'disposed',)

    def __init__(self, output):
        self.output = output
        self.state = StateMachine(self.STATE_GRAPH, self.STATE_NAMES)

        self.push_ev = Event()
        self.push_buf = deque()
        self.int_ev = Event()
        self.int_tag = object()
        self.shell_rdr = None

        def process(msg, dst, src):
            if self.shell_rdr is None:
                if msg is None:
                    self.int_ev(self.int_tag)
                    return False
                else:
                    op, args = msg
                    if op == self.OP_INT:
                        self.push_buf.appendleft(msg)
                        self.int_ev(self.int_tag)
                    elif op == self.OP_EXEC:
                        self.execute(*args)(lambda res: src.send(res))
                    else:
                        self.push_buf.append(msg)
                    self.push_ev(None)
                    return True
            else:
                self.shell_rdr.sender.send(msg, src)
                return True
        recv, send = pair()
        recv(process)

        ShellBase.__init__(self, send)

    @async_green
    def execute(self, locals, banner):
        """Execute shell
        """
        def raw_input(prompt=''):
            """Write a prompt and read a line
            """
            if prompt:
                self.output.write(prompt)
            while True:
                if self.push_buf:
                    op, args = self.push_buf.popleft()
                    if op == self.OP_PUSH:
                        return args
                    elif op == self.OP_INT:
                        raise KeyboardInterrupt()
                    elif op == self.OP_EOT:
                        raise EOFError()
                    raise RuntimeError('unknown prompt operation: {}'.format(op))
                bind_green(self.push_ev)

        def bind(target):
            """Get associated monadic value
            """
            value = bind_green(monad(target) | self.int_ev)
            if value is self.int_tag:
                if self.push_buf:  # dequeue excessive OP_INT
                    self.push_buf.popleft()
                raise KeyboardInterrupt()
            return value

        def bound(func):
            """Create version of asynchronous function
            """
            @functools.wraps(func)
            def func_bound(*args, **kwargs):
                return bind(func(*args, **kwargs))
            return func_bound

        def monad(target):
            """Get associated monad
            """
            monad_getter = getattr(target, '__monad__', None)
            if monad_getter is None:
                raise ValueError('target is not a monad')
            return monad_getter()

        @async
        def attach(conn, **locals):
            """Create shell on specified connection
            """
            yield self.redirect((yield conn(Shell)(self.output)), locals)

        try:
            self.state(self.STATE_EXEC)
            stderr_prev, sys.stderr = sys.stderr, self.output
            stdout_prev, sys.stdout = sys.stdout, self.output

            locals = locals or {}
            locals.update({
                'bind': bind,
                'bound': bound,
                'monad': monad,
                'fork': bound(ForkConnection),
                'ssh': bound(SSHConnection),
                'attach': bound(attach),
                'core': Core.local(),
            })
            console = code.InteractiveConsole(locals)
            for line in self.INIT_SOURCE.split('\n'):
                console.runsource(line.rstrip())
            console.raw_input = raw_input
            console.interact((self.BANNER if banner is None else banner).format(
                             host=socket.gethostname(), pid=os.getpid()))
        finally:
            sys.stderr = stderr_prev
            sys.stdout = stdout_prev
            self.state(self.STATE_DISP)
            self.sender.try_send(None)

    @async
    def redirect(self, shell, locals=None):
        """Redirect input events to a different shell object
        """
        if self.shell_rdr is not None:
            raise ValueError('has already been redirected')
        try:
            self.shell_rdr = shell
            yield shell(locals, banner='-> host:{host} pid:{pid}')
        finally:
            self.shell_rdr = None
            self.output.write('-> host:{} pid:{}\n'.format(socket.gethostname(), os.getpid()))


class ShellStream(object):
    """Send-able stream like object
    """
    def __init__(self, stream):
        encoding = getattr(stream, 'encoding', 'utf-8')
        get_stream = lazy(BufferedFile, stream, closefd=False)

        def write(data):
            stream = get_stream()
            stream.write_schedule(data.encode(encoding))
            stream.flush()()
            return True
        self.writer = Event()
        self.writer.on(write)

    def write(self, data):
        self.writer(data)
        return len(data)

    def flush(self):
        pass

    def isatty(self):
        return False

    def fileno(self):
        raise NotImplementedError()

    def close(self):
        self.writer.dispose()


@async
def prompt(shell, stdin=None):
    """Prompt shell
    """
    @async
    def prompt_coro():
        stream = None
        cancel = cancel_ev.future()
        try:
            stream = BufferedFile(stdin or sys.__stdin__, closefd=False)
            sigint = signal.signal(signal.SIGINT, lambda sig, stack:
                                   schedule()(lambda _: shell.interrupt()))
            while True:
                try:
                    line = yield async_any((stream.read_until_sub(b'\n'), cancel))
                    shell.push(line.rstrip().decode('utf-8'))
                except BrokenPipeError:
                    shell.end_of_transmission()
                except CanceledError:
                    break
        finally:
            signal.signal(signal.SIGINT, sigint)
            if stream is not None:
                stream.blocking(True)
                stream.dispose()
    try:
        cancel_ev = Event()
        yield async_any((shell(), prompt_coro()))
    finally:
        cancel_ev(Result.from_exception(CanceledError()))


if __name__ == '__main__':
    from ..app import app_run
    shell = Shell(ShellStream(sys.stdout))
    app_run(prompt(shell))



from . import stream, file, sock, sock_ssl, wrapped, buffered

from .stream import *
from .file import *
from .sock import *
from .sock_ssl import *
from .wrapped import *
from .buffered import *

__all__ = (stream.__all__ + file.__all__ + sock.__all__ + sock_ssl.__all__ +
           wrapped.__all__ + buffered.__all__)


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import tests

    suite = TestSuite()
    for test in (tests,):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite



"""Remoting module

Work with remote objects transparently.
"""
from . import hub, proxy, conn
from .hub import *
from .proxy import *
from .closure import *
from .conn import*

__all__ = hub.__all__ + proxy.__all__ + closure.__all__ + conn.__all__


def load_tests(loader, tests, pattern):
    """Load test protocol
    """
    from unittest import TestSuite
    from . import tests

    suite = TestSuite()
    for test in (tests,):
        suite.addTests(loader.loadTestsFromModule(test))

    return suite


def load_bench(runner):
    """Load benchmarks protocol
    """
    from . import bench
    bench.load_bench(runner)



import unittest
try:
    import greenlet
except ImportError:
    greenlet = None
from ..do_green import bind_green
from ..async import async_green
from ..result import Result
from ...event import Event

__all__ = ('DoGreenTest',)


class DoGreenTest(unittest.TestCase):
    @unittest.skipIf(greenlet is None, 'greenlet module is not installed')
    def test(self):
        event = Event()
        get = lambda: bind_green(event)
        ret = lambda val: rets.append(val)
        rets = []

        @async_green
        def green():
            return get()

        green()(ret)
        self.assertEqual(rets, [])

        event('value')
        self.assertEqual(rets.pop().value, 'value')

        green()(ret)
        event(Result.from_exception(ValueError('test')))
        with self.assertRaises(ValueError):
            rets.pop().value

        with self.assertRaises(RuntimeError):
            bind_green(event)  # bind outside of greenlet



"""Result monad (Haskell's Either monad)
"""
import os
import sys
import pdb
import socket
import textwrap
import traceback
import linecache
from .monad import Monad
from ..uniform import reraise, StringIO, PY2

__all__ = ('Result', 'result_excepthook', 'callsite_banner',)


class Result(Monad):
    """Result monad (Haskell's Either monad)
    """
    __slots__ = ('pair',)

    def __init__(self, pair):
        self.pair = pair

    @classmethod
    def from_value(cls, val):
        """From value factory
        """
        return cls.unit(val)

    @classmethod
    def from_error(cls, error):
        """From error factory
        """
        return cls((None, error))

    @classmethod
    def from_current_error(cls):
        """From current error factory
        """
        return cls((None, sys.exc_info()))

    @classmethod
    def from_exception(cls, exc):
        """From exception
        """
        try:
            raise exc
        except Exception:
            return cls.from_current_error()

    @property
    def error(self):
        return self.pair[1]

    @property
    def value(self):
        if self.pair[1] is None:
            return self.pair[0]
        else:
            reraise(*self.pair[1])

    def trace(self, debug=None, file=None, banner=None):
        """Show traceback

        Arguments:
            debug: weather start debugger on error or not
            file: output file object
            banner: callable object or None, if callable returns banner string
        """
        if self.pair[1] is not None:
            result_excepthook(*self.pair[1], file=file, banner=banner)
            if debug:
                pdb.post_mortem(self.pair[1][2])
        return self

    @classmethod
    def unit(cls, val):
        return cls((val, None))

    def bind(self, func):
        result, error = self.pair
        if error is None:
            try:
                return func(result)
            except Exception:
                return Result.from_current_error()
        else:
            return self

    def __eq__(self, other):
        if not isinstance(other, Result):
            return False
        if self.pair[1] is None:
            return self.pair == other.pair
        elif other.pair[1] is None:
            return False
        else:
            # compare only exception values in case of exception
            return self.pair[1][1] == other.pair[1][1]

    def __hash__(self):
        return hash(self.pair[0] if self.pair[1] is None else self.pair[1][1])

    def __reduce__(self):
        val, err = self.pair
        if err is None:
            return _from_value, (val,)
        else:
            tb = (self.tb_fmt.format(name=err[0].__name__, message=str(err[1]),
                                     tb=''.join(traceback.format_exception(*err))))
            tb += getattr(err[1], '_saved_traceback', '')
            exc = err[1]
            exc._saved_traceback = tb
            return _from_exc, (exc,)

    tb_fmt = textwrap.dedent("""\
        `-------------------------------------------------------------------------------
        Location : {host}/{pid}
        Error    : {{name}}: {{message}}

        {{tb}}""".format(host=socket.gethostname(), pid=os.getpid()))

    def __str__(self):
        return ('Result({})'.format(
                'val:{}'.format(self.pair[0]) if self.pair[1] is None else
                'err:{}'.format(repr(self.pair[1][1]))))

    def __repr__(self):
        return str(self)


def _from_value(val):
    return Result.from_value(val)


def _from_exc(exc):
    return Result.from_exception(exc)


def result_excepthook(et, eo, tb, file=None, banner=None):  # pragma: no cover
    """Result specific exception hook

    Correctly shows embedded traceback if any.

    Arguments:
        file: output file object
        banner: callable object or None, if callable returns banner string
    """
    stream = file or StringIO()
    if banner:
        stream.write(banner())
        stream.write('\n')

    tb = ''.join(traceback.format_exception(et, eo, tb))
    stream.write(tb.encode('utf-8') if PY2 else tb)

    tb_saved = getattr(eo, '_saved_traceback', None)
    if tb_saved is not None:
        stream.write(tb_saved)

    if file is None:
        sys.stderr.write(stream.getvalue())
        sys.stderr.flush()

# install result exception hook
sys.excepthook = result_excepthook


def callsite_banner(message=None, depth=2):
    """Call site banner

    Arguments:
        message: either string, or callable returning string
        depth: current frame depth from captured frame

    Returns banner function for current call site.
    """
    frame = sys._getframe(depth)
    code = frame.f_code
    lineno = frame.f_lineno
    globals = frame.f_globals
    del frame

    def banner():  # pragma: no cover
        filename = code.co_filename
        line = linecache.getline(filename, lineno, globals)
        if line:
            line = '\n    {}'.format(line.strip())
        if message is None:
            header = '[callsite] error caused by call'
        elif hasattr(message, '__call__'):
            header = message()
        else:
            header = message
        return textwrap.dedent("""\
            {header}
              File "{filename}", line {lineno}, in {name}{line}\
            """).format(header=header,
                        filename=filename,
                        lineno=lineno,
                        name=code.co_name,
                        line=line)
    return banner



"""Create pickle-able object closure from function
"""
import types
import marshal
import importlib
from ..uniform import PY2
__all__ = ('Closure',)


class Closure(object):
    """Pickle-able object which behaves as function
    """
    __slots__ = ('func',)

    def __init__(self, func):
        assert isinstance(func, types.FunctionType)
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __reduce__(self):
        closure = _closure_get(self.func)
        return (_closure_resotre,
               (marshal.dumps(self.func.__code__),
                self.func.__name__,
                self.func.__module__,
                self.func.__kwdefaults__ if not PY2 else self.func.func_defaults,
                None if closure is None else
                tuple(cell.cell_contents for cell in closure)))

    def __str__(self):
        return 'Closure({})'.format(self.func.__name__)

    def __repr__(self):
        return str(self)


def _closure_resotre(code, name, module, argdefs, closure):
    return Closure(types.FunctionType(
                   marshal.loads(code),
                   importlib.import_module(module).__dict__,
                   name,
                   argdefs,
                   None if closure is None else
                   tuple(_closure_get(lambda: val)[0] for val in closure)))


if hasattr(_closure_resotre, '__closure__'):
    def _closure_get(func):
        return func.__closure__
else:
    def _closure_get(func):
        return func.func_closure



"""Manipulate system processes asynchronously
"""
import io
import os
import sys
import errno
import pickle
import signal
import atexit
from .event import Event
from .dispose import CompDisp, FuncDisp
from .monad import Result, Cont, async, async_all, do_return, do_done
from .core import Core
from .stream import BufferedFile, fd_close_on_exec
from .uniform import BrokenPipeError
from .state_machine import StateMachine

__all__ = ('Process', 'ProcessError', 'ProcessPipe', 'PIPE', 'DEVNULL',
           'process_call', 'process_chain_call',)


def devnull_fd():
    """Get devnull file descriptor
    """
    global DEVNULL_FD
    if DEVNULL_FD is None:
        DEVNULL_FD = os.open(os.devnull, os.O_RDWR)
    return DEVNULL_FD


@atexit.register
def devnull_fd_dispose():  # pragma: no cover
    """Close devnull file descriptor
    """
    global DEVNULL_FD
    fd, DEVNULL_FD = DEVNULL_FD, None
    if fd is not None:
        os.close(fd)


PIPE = -1
DEVNULL = -2
DEVNULL_FD = None
STDIN_FD = sys.__stdin__.fileno() if sys.__stdin__ else devnull_fd()
STDOUT_FD = sys.__stdout__.fileno() if sys.__stdout__ else devnull_fd()
STDERR_FD = sys.__stderr__.fileno() if sys.__stderr__ else devnull_fd()


class ProcessError(Exception):
    """Process specific error
    """


class Process(object):
    """Asynchronous child process
    """
    STATE_INIT = 0
    STATE_FORK = 1
    STATE_RUN = 2
    STATE_DISP = 3
    STATE_GRAPH = StateMachine.compile_graph({
        STATE_INIT: (STATE_DISP, STATE_FORK,),
        STATE_FORK: (STATE_DISP, STATE_RUN,),
        STATE_RUN:  (STATE_DISP,),
        STATE_DISP: (STATE_DISP,),
    })
    STATE_NAMES = ('not-run', 'forking', 'running', 'disposed',)

    def __init__(self, command, stdin=None, stdout=None, stderr=None,
                 preexec=None, shell=None, environ=None, check=None,
                 bufsize=None, kill=None, core=None):
        # options
        class options(object):
            def __getattr__(self, name):
                return opts[name]
        if isinstance(command, str):
            command = [command]
        opts = {
            'stdin': stdin,
            'stdout': stdout,
            'stderr': stderr,
            'environ': environ,
            'command': ['/usr/bin/env', 'sh', '-c', ' '.join(command)] if shell else command,
            'kill': kill or -1,  # don't kill process by default
            'check': check is None or check,
            'bufsize': bufsize,
            'preexec': preexec,
            'status': Event(),
        }
        self.opts = options()

        # properties
        self.pid = None
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self.core = core or Core.local()
        self.state = StateMachine(self.STATE_GRAPH, self.STATE_NAMES)
        self.status = self.opts.status.future()

        # killing
        def schedule_kill():
            if(not self.pid or
               self.status.completed or
               self.opts.kill < 0):
                return

            def kill(_):
                if not self.status.completed:
                    try:
                        os.killpg(self.pid, signal.SIGTERM)
                    except OSError as error:
                        if error.errno != errno.ECHILD:
                            self.opts.status(Result.from_current_error())
            self.core.sleep(self.opts.kill)(kill)

        # disposal
        self.dispose = CompDisp()
        self.dispose.add_action(schedule_kill)
        self.dispose.add_action(lambda: self.state(self.STATE_DISP))
        if self.opts.kill >= 0:
            self.core.dispose.add(self)
            self.dispose.add_action(lambda: self.core.dispose.remove(self))

    @async
    def __call__(self):
        self.state(self.STATE_FORK)
        core = self.core
        try:
            def pipe(file, default, reader):
                """Create pair of getters for stream and associated descriptor
                """
                def fake_pipe(child_fd):
                    def detach(fd=None):
                        if pid == os.getpid():
                            assert fd is None, 'fd option must not be used in parent process'
                            return None
                        else:  # pragma: no cover (child process)
                            if fd is None or fd == child_fd:
                                return child_fd
                            else:
                                os.dup2(child_fd, fd)
                                os.close(child_fd)
                                return fd
                    return detach
                pid = os.getpid()

                # from default descriptor
                if file is None:
                    return fake_pipe(default)
                # from pipe
                elif file == PIPE:
                    return self.dispose.add(ProcessPipe(reader, bufsize=self.opts.bufsize, core=core))
                # from /dev/null
                elif file == DEVNULL:
                    return fake_pipe(devnull_fd())
                # from custom descriptor
                else:
                    return fake_pipe(file if isinstance(file, int) else file.fileno())

            stdin = pipe(self.opts.stdin, STDIN_FD, False)
            stdout = pipe(self.opts.stdout, STDOUT_FD, True)
            stderr = pipe(self.opts.stderr, STDERR_FD, True)
            status_pipe = self.dispose.add(ProcessPipe(True, bufsize=self.opts.bufsize, core=core))

            self.pid = os.fork()
            # parent
            if self.pid:
                def schedule_dispose(stream):
                    if stream is None:
                        return
                    return self.dispose.add(stream)
                self.stdin = schedule_dispose(stdin())
                self.stdout = schedule_dispose(stdout())
                self.stderr = schedule_dispose(stderr())

                waitpid = self.core.waitpid(self.pid).future()  # no zombies
                try:
                    error_data = yield self.dispose.add(status_pipe()).read_until_eof()
                    if error_data:
                        pickle.loads(error_data).value  # will raise captured error
                except BrokenPipeError:
                    pass
                self.state(self.STATE_RUN)

                @async
                def check_status():
                    status = yield waitpid
                    if self.opts.check and status != 0:
                        raise ProcessError('command {} returned non-zero exit status {}'
                                           .format(self.opts.command, status))
                    do_return(status)
                check_status()(self.opts.status)

            # child
            else:  # pragma: no cover
                try:
                    status_fd = status_pipe()
                    fd_close_on_exec(status_fd, True)
                    stdin(0)
                    stdout(1)
                    stderr(2)
                    if self.opts.kill > 0:
                        os.setpgrp()
                    if self.opts.preexec is not None:
                        self.opts.preexec()
                    os.execvpe(self.opts.command[0], self.opts.command,
                               self.opts.environ or os.environ)
                except Exception:
                    with io.open(status_fd, 'wb') as error_stream:
                        pickle.dump(Result.from_current_error(), error_stream)
                finally:
                    getattr(os, '_exit', lambda _: os.kill(os.getpid(), signal.SIGKILL))(111)

            do_return(self)
        except Exception:
            self.opts.status(Result.from_current_error())
            self.dispose()
            raise

    def __monad__(self):
        return self()

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        """String representation of the process
        """
        if self.status.completed:
            status = ', status:{}'.format(self.status.res)
        else:
            status = ''
        return ('Process(state:{}, pid:{}, cmd:{}{})'.format(self.state.state_name(),
                self.pid, self.opts.command, status))

    def __repr__(self):
        return str(self)


class ProcessPipe(object):
    """Pipe to communicate with forked process
    """
    def __init__(self, reader, bufsize=None, core=None):
        self.core = core
        self.bufsize = bufsize
        self.parent_pid = os.getpid()
        self.parent_fd, self.child_fd = os.pipe() if reader else reversed(os.pipe())
        fd_close_on_exec(self.parent_fd, True)

    def __call__(self, fd=None):
        """Detach appropriate object

        Detaches stream in parent process, and descriptor in child process.
        """
        parent_pid, self.parent_pid = self.parent_pid, None
        if parent_pid is None:
            raise RuntimeError('pipe has already been detached')
        if parent_pid == os.getpid():
            assert fd is None, 'fd option must not be used in parent process'
            os.close(self.child_fd)
            stream = BufferedFile(self.parent_fd, bufsize=self.bufsize, core=self.core)
            return stream
        else:  # pragma: no cover (child process)
            os.close(self.parent_fd)
            if fd is None or fd == self.child_fd:
                return self.child_fd
            else:
                os.dup2(self.child_fd, fd)
                os.close(self.child_fd)
                return fd

    def dispose(self):
        parent_pid, self.parent_pid = self.parent_pid, None
        if parent_pid is not None:
            os.close(self.parent_fd)
            os.close(self.child_fd)

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return ('ProcessPipe(pid:{}, parent_fd:{}, child_fd:{})'.format
               (self.parent_pid, self.parent_fd, self.child_fd))

    def __repr__(self):
        return str(self)


@async
def process_call(command, stdin=None, stdout=None, stderr=None,
                 preexec=None, shell=None, environ=None, check=None,
                 bufsize=None, kill=None, core=None):
    """Run command

    Returns:
        (output, error, code) where output and error bytes or None and code
        is a process status code.
    """
    stdin_data = None
    if stdin is None:
        stdin = PIPE
    elif isinstance(stdin, (bytes, memoryview)):
        stdin, stdin_data = PIPE, stdin
    stdout = PIPE if stdout is None else stdout
    stderr = PIPE if stderr is None else stderr

    with (yield Process(command=command, stdin=stdin, stdout=stdout,
                        stderr=stderr, preexec=preexec, shell=shell,
                        environ=environ, check=check, bufsize=bufsize,
                        kill=kill, core=core)) as proc:
        in_cont = Cont.unit(None)
        if stdin_data:
            proc.stdin.write_schedule(stdin_data)
            in_cont = proc.stdin.flush_and_dispose()
        elif proc.stdin:
            proc.stdin.dispose()

        out_cont = proc.stdout.read_until_eof() if proc.stdout else Cont.unit(None)
        err_cont = proc.stderr.read_until_eof() if proc.stderr else Cont.unit(None)

        _, out, err, status = yield async_all((in_cont, out_cont, err_cont, proc.status))
        do_return((out, err, status))


@async
def process_chain_call(commands, stdin=None, stdout=None, stderr=None, **proc_opts):
    """Run pipe chained commands

    Returns:
        (output, error, codes) where output and error bytes or None and codes
        is a list of process status codes.
    """
    if not commands:
        raise ValueError('commands list must not be empty')

    with CompDisp() as dispose:
        stdin_data = None
        if stdin is None:
            stdin = PIPE
        elif isinstance(stdin, (bytes, memoryview)):
            stdin, stdin_data = PIPE, stdin
        stdout = PIPE if stdout is None else stdout

        # standard error stream (common for all commands in pipe)
        err_cont = Cont.unit(None)
        if stderr in (PIPE, None):
            stderr_fd, stderr = os.pipe()
            stderr_stream = dispose.add(BufferedFile(stderr_fd,
                                                     bufsize=proc_opts.get('bufsize'),
                                                     core=proc_opts.get('core')))
            stderr_stream.close_on_exec(True)
            stderr_close = dispose.add(FuncDisp(lambda: os.close(stderr)))

            @async
            def err_coro():
                stderr_close()
                do_done(stderr_stream.read_until_eof())
            err_cont = err_coro()

        # start processes
        def disp_fd(fd):
            """Register descriptor for disposal

            Needed to correctly capture descriptor inside loop.
            """
            dispose.add(FuncDisp(lambda: os.close(fd)))
            return fd

        procs = []
        for command in commands[:-1]:
            proc = yield dispose.add(Process(command, stdin=stdin, stdout=PIPE,
                                             stderr=stderr, **proc_opts))
            procs.append(proc)
            stdin = disp_fd(proc.stdout.detach())
        proc = yield dispose.add(Process(commands[-1], stdin=stdin, stdout=stdout,
                                         stderr=stderr, **proc_opts))
        procs.append(proc)

        # write standard input
        in_cont = Cont.unit(None)
        proc_first = procs[0]
        if stdin_data:
            proc_first.stdin.write_schedule(stdin_data)
            in_cont = proc_first.stdin.flush_and_dispose()
        elif proc_first.stdin:
            proc_first.stdin.dispose()

        # receive standard output and wait for processes completion
        out_cont = proc.stdout.read_until_eof() if proc.stdout else Cont.unit(None)
        _, out, err, status = yield async_all((in_cont, out_cont, err_cont,
                                               async_all(proc.status for proc in procs)))
        do_return((out, err, status))



from collections import deque

__all__ = ('Event', 'EventQueue',)


class Event(object):
    """Event object
    """
    __slots__ = ('handlers',)

    def __init__(self):
        self.handlers = []

    def __call__(self, event):
        """Fire event
        """
        handlers, self.handlers = self.handlers, []
        for handler in handlers:
            if handler(event):
                self.handlers.append(handler)

    def on(self, handler):
        """On handler

        Returns specified handler. If handler returned value is False it will
        be automatically unsubscripted.
        """
        self.handlers.append(handler)
        return handler

    def on_once(self, handler):
        """Execute handler only once
        """
        def once_handler(event):
            handler(event)
            return False
        self.on(once_handler)

    def __iadd__(self, handler):
        self.on(handler)
        return self

    def off(self, handler):
        """Off handler

        Preferred way to do un-subscription is to return False from handler.
        Returns True in case of successful un-subscription.
        """
        try:
            self.handlers.remove(handler)
            return True
        except ValueError:
            return False

    def __isub__(self, handler):
        self.off(handler)
        return self

    def __monad__(self):
        """Continuation for nearest event
        """
        return Cont(lambda ret: self.on_once(ret))

    def future(self):
        """Future for nearest event
        """
        return self.__monad__().future()

    def __str__(self):
        return '{}(handlers:{})'.format(type(self).__name__, len(self.handlers))

    def __repr__(self):
        return str(self)

    def __reduce__(self):
        return ReducedEvent, (self.reduce().sender,)

    def reduce(self, hub=None):
        """Create reduced event

        Fire-able and send-able over remote connection but not subscribe-able.
        """
        from .remoting import pair

        def fire_event(msg, dst, src):
            op, event = msg
            if op == EVENT_FIRE:
                self(event)
                return True
            elif op == EVENT_DISPOSE:
                return False
            else:
                raise ValueError('unknown reduced event operation')
        recv, send = pair(hub=hub)
        recv(fire_event)
        return ReducedEvent(send)

    def dispose(self):
        """Does nothing

        Compatibility with reduced event interface.
        """

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False


class EventQueue(Event):
    """Event queue

    Event object which queues event to be handled later if no handler is set.
    """
    __slots__ = Event.__slots__ + ('queue',)

    def __init__(self):
        Event.__init__(self)
        self.queue = deque()

    def __call__(self, event):
        if self.handlers:
            Event.__call__(self, event)
        else:
            self.queue.append(event)

    def on(self, handler):
        Event.on(self, handler)
        while self.handlers:
            if not self.queue:
                return
            Event.__call__(self, self.queue.popleft())

    def __len__(self):
        return len(self.queue)

    def __bool__(self):
        return bool(self.queue)

    def __str__(self):
        return '{}(len:{}, handlers:{})'.format(type(self).__name__,
                                                len(self.queue), len(self.handlers))


class ReducedEvent(object):
    """Reduced event

    Can be called and send over remote connection but cannot be subscribed to.
    """
    __slots__ = ('sender',)

    def __init__(self, sender):
        self.sender = sender

    def __call__(self, event):
        if self.sender is not None:
            if self.sender.try_send((EVENT_FIRE, event)):
                return
            self.dispose()
        raise ValueError('reduced event is disposed')

    def dispose(self):
        sender, self.sender = self.sender, None
        if sender:
            sender.try_send((EVENT_DISPOSE, None))

    def __enter__(self):
        return self

    def __exit__(self, et, eo, tb):
        self.dispose()
        return False

    def __str__(self):
        return 'ReducedEvent(addr:{})'.format(self.sender.addr
                                              if self.sender else None)

    def __repr__(self):
        return str(self)

EVENT_FIRE = 0
EVENT_DISPOSE = 1

# Event object is used inside async_single, defer import of Cont
from .monad.cont import Cont



"""Wrapped stream
"""
from .stream import Stream
from ..monad import async, do_return

__all__ = ('WrappedStream',)


class WrappedStream(Stream):
    """Wrapped stream

    Stream wrapped around base stream.
    """
    def __init__(self, base):
        Stream.__init__(self)
        self.initing()
        self.base = base

    def __getattr__(self, name):
        if self.disposed:
            raise ValueError('stream is disposed')
        return getattr(self.base, name)

    def fileno(self):
        if self.disposed:
            raise ValueError('stream is disposed')
        return self.base.fileno()

    @async
    def read(self, size):
        with self.reading:
            do_return((yield self.base.read(size)))

    @async
    def write(self, data):
        with self.writing:
            do_return((yield self.base.write(data)))

    def flush(self):
        if self.disposed:
            raise ValueError('stream is disposed')
        return self.base.flush()

    def dispose(self):
        if Stream.dispose(self):
            if self.base is not None:
                self.base.dispose()
            return True
        return False

    def detach(self):
        base, self.base = self.base, None
        if not self.dispose():
            raise ValueError('stream is disposed')
        return base

    def __str__(self):
        return ('{}({})'.format(type(self).__name__, self.base))
